"""
This module will create an amber mdin file for either sander or
pmemd (or others). The program specification loads the appropriate
dictionaries with default values, etc. It can read and write mdins.
"""
from io import TextIOBase
from re import findall

# This module will create and read a sander/pmemd input
from .cntrl import cntrl
from .ewald import ewald
from .pb import pb
from .qmmm import qmmm
from .rism import rism
from .gbnsr6 import gbnsr6
from ...exceptions import InputError


def addOn(line, string, file):
    if len(line.strip()) == 0:
        return line + string
    elif len(line) + len(string) > 40:
        file.write(line + '\n')
        return ' ' + string
    else:
        return line + string

def splitStr(string):
    # split string at commas not within single or double quotes
    pattern = r"""('[^']*'|"[^"]*"|[^'",\s]+)"""
    matches = findall(pattern, string)

    return matches


class Mdin:

    def __init__(self, program = 'sander', verbosity = 1):
        # define instance data
        self.program = program   # which program we're creating the input file for
        self.cntrl_obj = cntrl() # object with cntrl namelist vars in a dictionary
        self.ewald_obj = ewald() # object with ewald namelist vars in a dictionary
        self.pb_obj = pb()       # object with pb namelist vars in a dictionary
        self.qmmm_obj = qmmm()   # object with qmmm namelist vars in a dictionary
        self.rism_obj = rism()   # object with rism namelist vars in a dictionary
        self.gbnsr6_obj = gbnsr6()   # object with gbnsr6 namelist vars in a dictionary
        self.verbosity = 0       # verbosity level: 0 -- print nothing
                                 #                  1 -- print errors
                                 #                  2 -- 1 + warnings
                                 #                  3 -- 2 + notes
        self.cards = []          # array that has all of the input cards that come
                                 # after namelists
        self.cntrl_nml = {}      # dictionary with cntrl namelist vars
        self.cntrl_nml_defaults = {} # dictionary with default cntrl namelist vars
        self.ewald_nml = {}          # dictionary with ewald namelist vars
        self.ewald_nml_defaults = {} # dictionary with default ewald namelist vars
        self.pb_nml = {}             # dictionary with pb namelist vars
        self.pb_nml_defaults = {}    # dictionary with default pb namelist vars
        self.qmmm_nml = {}           # dictionary with qmmm namelist vars
        self.qmmm_nml_defaults = {}  # dictionary with default qmmm namelist vars
        self.rism_nml = {}           # dictionary with rism namelist vars
        self.rism_nml_defaults = {}  # dictionary with default rism namelist vars
        self.gbnsr6_nml = {}           # dictionary with gbnsr6 namelist vars
        self.gbnsr6_nml_defaults = {}  # dictionary with default gbnsr6 namelist vars
        self.valid_namelists = {}    # array with valid namelists for each program
        self.title = 'mdin prepared by ParmEd'   # title for the mdin file


        if self.program == "sander":
            self.cntrl_nml = self.cntrl_obj.sander
            self.ewald_nml = self.ewald_obj.sander
            self.pb_nml = self.pb_obj.sander
            self.qmmm_nml = self.qmmm_obj.sander
            self.rism_nml = self.rism_obj.sander
            self.valid_namelists = {'cntrl', 'ewald', 'qmmm', 'pb', 'rism'}
        elif self.program == "sander.APBS":
            self.cntrl_nml = self.cntrl_obj.sander
            self.pb_nml = self.pb_obj.sanderAPBS
            self.valid_namelists = {'cntrl', 'apbs'}
        elif self.program == "pmemd":
            self.cntrl_nml = self.cntrl_obj.pmemd
            self.ewald_nml = self.ewald_obj.pmemd
            self.valid_namelists = {'cntrl', 'ewald'}
        elif self.program == "gbnsr6":
            self.cntrl_nml = self.cntrl_obj.gbnsr6
            self.gbnsr6_nml = self.gbnsr6_obj.gbnsr6
            self.valid_namelists = ['cntrl', 'gb']
        else:
            raise InputError(f"Unrecognized program [{self.program}]")

        self.cntrl_nml_defaults = self.cntrl_nml.copy()
        self.ewald_nml_defaults = self.ewald_nml.copy()
        self.pb_nml_defaults = self.pb_nml.copy()
        self.qmmm_nml_defaults = self.qmmm_nml.copy()
        self.rism_nml_defaults = self.rism_nml.copy()
        self.gbnsr6_nml_defaults = self.gbnsr6_nml.copy()

    def write(self, filename: str = 'mdin'):

        def write_nml(nml, defaults, file: TextIOBase, header, print_header=False) -> None:
            # automatic indent of single space
            line = ' '
            header_printed = False
            # add any variable that is different from the default to the mdin file
            for var, val in nml.items():
                if val == defaults[var]:
                    continue
                if not header_printed:
                    file.write(f"{header}\n")
                    header_printed = True
                if isinstance(val, (list, tuple)):
                    if isinstance(val[0], str):
                        temp_out = ','.join(map(lambda x: f"'{x}'", val))
                        line = addOn(line, f"{var}={temp_out}, ", file)
                    else:
                        line = addOn(line, f"{var}={','.join(map(str, val))}, ", file)
                elif isinstance(val, str):
                    line = addOn(line, f"{var}='{val}', ", file)
                else:
                    line = addOn(line, f"{var}={val}, ", file)
            # flush any remaining items that haven't yet been printed to the mdin file
            if line.strip():
                file.write(f"{line}\n")

            # gbnsr6 requieres always print the cntrl header, so we check if was already printed
            if not header_printed and print_header:
                file.write(f"{header}\n")
                header_printed = True

            # End namelist
            if header_printed:
                file.write('/\n')

        # open the file for writing and write the header and &cntrl namelist
        file = open(filename,'w')
        file.write(self.title + '\n')
        write_nml(self.cntrl_nml, self.cntrl_nml_defaults, file, "&cntrl", True)
        write_nml(self.ewald_nml, self.ewald_nml_defaults, file, "&ewald")
        pb_nml_name = "&apbs" if self.program == "sander.APBS" else "&pb"
        write_nml(self.pb_nml, self.pb_nml_defaults, file, pb_nml_name)
        write_nml(self.gbnsr6_nml, self.gbnsr6_nml_defaults, file, "&gb")
        if self.cntrl_nml.get('irism'):
            write_nml(self.rism_nml, self.rism_nml_defaults, file, "&rism", True)
        if self.cntrl_nml.get("ifqnt") == 1:
            write_nml(self.qmmm_nml, self.qmmm_nml_defaults, file, "&qmmm")

        # Write the cards to the input file
        file.write("\n".join([card.strip() for card in self.cards]))
        if len(self.cards) != 0:
            file.write('\nEND\n')

        file.close()

    def read(self, filename = 'mdin'):
        lines = open(filename, 'r').readlines()

        # split up input file into separate fields by comma
        blocks = []  # namelists in the order they appear
        block_fields = [] # array of arrays that correspond to entries in
                          # namelists found in "blocks" above
        inblock = False
        lead_comment = True
        for i in range(len(lines)):
            if not inblock and not lines[i].strip().startswith('&') and lead_comment:
                continue
            elif not inblock and not lines[i].strip().startswith('&') and not lead_comment:
                final_ended = True
                for j in range(i,len(lines)):
                    if lines[j].strip().startswith('&'):
                        final_ended = False
                if final_ended and len(lines[i].strip()) != 0:
                    self.cards.append(lines[i])
            elif not inblock and lines[i].strip().startswith('&'):
                lead_comment = False
                inblock = True
                block = lines[i].strip()[1:].lower()
                blocks.append(block)    # add the name of the namelist to "blocks"
                block_fields.append([]) # add empty array to be filled with entries
                                        # for given namelist
                if not block in self.valid_namelists:
                    raise InputError(f'Invalid namelist ({lines[i].strip()}) in input file ({filename}) for {self.program}')
            elif inblock and (lines[i].strip() == '/' or lines[i].strip() == '&end'):
                inblock = False
            elif inblock and lines[i].strip().startswith('&'):
                raise InputError(f"Invalid input file ({filename}). Namelist not terminated")
            elif inblock:
                items = splitStr(lines[i].strip())
                j = 0
                while j < len(items):
                    items[j] = items[j].strip()
                    if len(items[j]) == 0:
                        items.pop(j)
                    else:
                        j += 1
                block_fields[len(block_fields)-1].extend(items)

        # take out the last END in the cards if it's there
        if len(self.cards) != 0 and self.cards[len(self.cards)-1].strip().upper() == 'END':
            self.cards.pop()

        # combine any multi-element fields: e.g. rstwt=1,2,3,
        begin_field = -1
        for i in range(len(block_fields)):
            for j in range(len(block_fields[i])):
                if block_fields[i][j].startswith("'") or block_fields[i][j].startswith("\"") or not '=' in block_fields[i][j]:
                    if begin_field == -1:
                        raise InputError(f'Invalid input file ({filename}).')
                    else:
                        if block_fields[i][j].startswith("'") or block_fields[i][j].startswith("\""):
                            block_fields[i][begin_field] += block_fields[i][j]
                        
                            # prevents treating '=' in quoted field values (e.g. wildcards for restraintmask)
                            # from delimiting new fields
                            if '=' in block_fields[i][j]:
                                block_fields[i][j] = block_fields[i][j].replace('=','')
                        
                        else:
                            block_fields[i][begin_field] += ',' + block_fields[i][j]
                else:
                    begin_field = j

        # now parse through the options and add them to the dictionaries
        for i in range(len(block_fields)):
            for j in range(len(block_fields[i])):
                if not '=' in block_fields[i][j]:
                    continue
                else:
                    var = block_fields[i][j].split('=', 1)
                    self.change(blocks[i], var[0].strip(), var[1].strip())

    def change(self, namelist, variable, value):
        """ Change the value of a variable without adding a new key-pair """

        namelist_map = dict(
            cntrl=(self.cntrl_nml, self.cntrl_nml_defaults),
            ewald=(self.ewald_nml, self.ewald_nml_defaults),
            pb=(self.pb_nml, self.pb_nml_defaults),
            apbs=(self.pb_nml, self.pb_nml_defaults),
            qmmm=(self.qmmm_nml, self.qmmm_nml_defaults),
            rism=(self.rism_nml, self.rism_nml_defaults),
            gb=(self.gbnsr6_nml, self.gbnsr6_nml_defaults)
        )

        variable = variable.lower()
        if isinstance(value, str) and len(value) > 0:
            if (value[0] == value[-1] == "'") or (value[0] == value[-1] == '"'):
                value = value[1:-1]

        if namelist in namelist_map:

            nml, nml_defaults = namelist_map[namelist]
            if variable in nml:
                mytype = type(nml_defaults[variable])
                nml[variable] = mytype(value)
            else:
                raise InputError(f"Unknown variable {variable} in &{namelist}")
        else:
            raise InputError(f'Unknown namelist ({namelist})!')

    def check(self):
        return True

    def SHAKE(self):
        self.change('cntrl','ntf', 2)
        self.change('cntrl','ntc', 2)
        self.change('cntrl','dt', 0.002)

    def constPressure(self, press=1.0, taup=1.0):
        self.change('cntrl','ntb', 2)
        self.change('cntrl','ntp', 1)
        self.change('cntrl','pres0', press)
        self.change('cntrl','taup', taup)

    def constVolume(self):
        self.change('cntrl','ntb', 1)
        self.change('cntrl','ntp', 0)

    def constTemp(self, ntt=3, temp=300.0, gamma_ln=2.0, ig=-1, tautp=1.0):
        self.change('cntrl','ntt', ntt)
        self.change('cntrl','temp0', temp)
        self.change('cntrl','tempi', temp)
        self.change('cntrl','gamma_ln', gamma_ln if ntt==3 else 0)
        self.change('cntrl','ig', ig)
        self.change('cntrl','tautp', tautp)

    def constpH(self, solvph=7.0, igb=2, ntcnstph=10):
        self.change('cntrl','icnstph', 1)
        self.change('cntrl','solvph', solvph)
        self.change('cntrl','ntcnstph', ntcnstph)
        self.change('cntrl','igb', igb)
        self.change('cntrl','ntb', 0)
        self.change('cntrl','saltcon', 0.1)

    def restrainHeavyAtoms(self, restraint_wt=0.0):
        self.change('cntrl','ntr', 1)
        self.change('cntrl','restraint_wt', restraint_wt)
        self.change('cntrl','restraintmask', '!@H=')

    def restrainBackbone(self, restraint_wt=0.0):
        self.change('cntrl','ntr', 1)
        self.change('cntrl','restraint_wt', restraint_wt)
        self.change('cntrl','restraintmask', '@N,CA,C')

    def genBorn(self, igb=5, rgbmax=25.0):
        self.change('cntrl','igb', igb)
        self.change('cntrl','ntb', 0)
        self.change('cntrl','ntp', 0)
        self.change('cntrl','rgbmax', rgbmax)

    def time(self, time=1000.0, dt=None): # time in ps
        if dt is None:
            if self.cntrl_nml['ntc'] == 2 and self.cntrl_nml['ntf'] == 2:
                dt = 0.002
            else:
                dt = 0.001
        time = int(time / dt)

        self.change('cntrl','dt', dt)
        self.change('cntrl','nstlim', time)
        self.change('cntrl','imin', 0)

    def heat(self, tempi=0.0, temp0=300.0, ntt=3, tautp=5.0, ig=-1, gamma_ln=5.0):
        self.constVolume()
        self.change('cntrl','tempi', tempi)
        self.change('cntrl','temp0', temp0)
        self.change('cntrl','ntt', ntt)
        self.change('cntrl','tautp', tautp)
        self.change('cntrl','ig', ig)
        self.change('cntrl','gamma_ln', gamma_ln if ntt==3 else 0)

    def restart(self,ntx=5):
        self.change('cntrl','irest',1)
        self.change('cntrl','ntx',ntx)

    def TI(self, clambda=0.0):
        self.change('cntrl','clambda', clambda)
        self.change('cntrl','icfe',1)

    def softcore_TI(self, scalpha=0.5, scmask='', crgmask='', logdvdl=0):
        self.change('cntrl','icfe',1)
        self.change('cntrl','ifsc',1)
        self.change('cntrl','scalpha',scalpha)
        self.change('cntrl','scmask',scmask)
        self.change('cntrl','crgmask',crgmask)
        self.change('cntrl','logdvdl',logdvdl)

    def minimization(self, imin=1, maxcyc=1, ncyc=10, ntmin=1):
        self.change('cntrl','imin', imin)
        self.change('cntrl','maxcyc', maxcyc)
        self.change('cntrl','ncyc', ncyc)
        self.change('cntrl','ntmin', ntmin)

    def rism(self, imin=5, irism=1):
        self.change('cntrl', 'imin', imin)
        self.change('cntrl', 'irism', irism)

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def AddCard(self, title='Residues in card', cardString='RES 1'):
        self.cards.append('%s\n%s\nEND' % (title, cardString))



#!/usr/bin/env python
from __future__ import division, print_function

import sys

# OpenMM Imports
import openmm as mm
import openmm.app as app

# ParmEd Imports
from parmed import load_file, unit as u
from parmed.charmm import CharmmParameterSet
from parmed.openmm import StateDataReporter

# Load the CHARMM files
print('Loading CHARMM files...')
params = CharmmParameterSet('toppar/par_all36_prot.prm')
ala5_gas = load_file('ala5_autopsf.psf')
ala5_crds = load_file('ala5_autopsf.pdb')

# NOTE NOTE
# The parameter set we used here is the CHARMM 36 force field, but this is
# strictly an example. It is important that you use the most accurate (typically
# most up-to-date) force fields for your own simulation. See the CHARMM
# parameter web page for updates:
# http://mackerell.umaryland.edu/CHARMM_ff_params.html
# END NOTE

# Create the OpenMM system
print('Creating OpenMM System')
system = ala5_gas.createSystem(params, nonbondedMethod=app.NoCutoff,
                               constraints=app.HBonds, implicitSolvent=app.HCT,
                               implicitSolventSaltConc=0.1*u.moles/u.liter,
)

# Create the integrator to do Langevin dynamics
integrator = mm.LangevinIntegrator(
                        300*u.kelvin,       # Temperature of heat bath
                        1.0/u.picoseconds,  # Friction coefficient
                        2.0*u.femtoseconds, # Time step
)

# Define the platform to use; CUDA, OpenCL, CPU, or Reference. Or do not specify
# the platform to use the default (fastest) platform
platform = mm.Platform.getPlatformByName('CUDA')
prop = dict(CudaPrecision='mixed') # Use mixed single/double precision

# Create the Simulation object
sim = app.Simulation(ala5_gas.topology, system, integrator, platform, prop)

# Set the particle positions
sim.context.setPositions(ala5_crds.positions)

# Minimize the energy
print('Minimizing energy')
sim.minimizeEnergy(maxIterations=500)

# Set up the reporters to report energies and coordinates every 100 steps
sim.reporters.append(
        StateDataReporter(sys.stdout, 100, step=True, potentialEnergy=True,
                          kineticEnergy=True, temperature=True,
                          volume=True, density=True)
)
sim.reporters.append(app.DCDReporter('ala5_gb.dcd', 100))

# Run dynamics
print('Running dynamics')
sim.step(10000)



from argparse import ArgumentParser
import datetime
import os
try:
    import readline
except ImportError:
    readline = None
import signal
import sys
import warnings
from . import load_file
from .exceptions import ParmedError
from .tools.logos import Logo
from .tools.exceptions import SeriousParmWarning, InterpreterError, ParmError
from .tools.parmed_cmd import ParmedCmd
from .tools.actions import Action
from .tools.parmlist import ParmList
from . import __version__

def clapp():
    # Set up new excepthook to clean up fatal exception printouts
    def interrupted(*args, **kwargs):
        """ Handle interruptions gracefully """
        sys.stdout.write('Interrupted\n')
        sys.exit(1)

    signal.signal(signal.SIGINT, interrupted)

    # Define our own custom warning printer
    def _print_warnings(message, category, filename, lineno, file=None, line=None):
        """ Override the default showwarning method """
        file = sys.stderr if file is None else file
        try:
            file.write(f'{category.__name__}: {message}\n')
        except IOError:
            pass

    warnings.showwarning = _print_warnings

    # Set up parser
    parser = ArgumentParser()
    parser.add_argument('-v', '--version', action='version', version=f"%(prog)s: Version {__version__}")
    group = parser.add_argument_group('Input Files')
    group.add_argument('-i', '--input', dest='script', default=[],
             metavar='FILE', help='''Script with ParmEd commands to execute. Default
             reads from stdin. Can be specified multiple times to process multiple
             input files.''', action='append')
    group.add_argument('-p', '--parm', dest='prmtop', default=[],
             metavar='<prmtop>', action='append', help='''List of topology files to load into
             ParmEd. Can be specified multiple times to process multiple topologies.''')
    group.add_argument('-c', '--inpcrd', dest='inpcrd', default=[],
             metavar='<inpcrd>', action='append', help='''List of inpcrd files to
             load into ParmEd. They are paired with the topology files in the same
             order that each set of files is specified on the command-line.''')
    group = parser.add_argument_group('Output Files')
    group.add_argument('-O', '--overwrite', dest='overwrite', default=False,
             help='Allow ParmEd to overwrite existing files.', action='store_true')
    group.add_argument('-l', '--logfile', dest='logfile', metavar='FILE',
             default='parmed.log', help='''Log file with every command executed
             during an interactive ParmEd session. Default is parmed.log''')
    group = parser.add_argument_group('Interpreter Options', '''These options affect
             how the ParmEd interpreter behaves in certain cases.''')
    group.add_argument('--prompt', dest='prompt', default='>', metavar='PROMPT',
             help='String to use as a command prompt.')
    group.add_argument('-n', '--no-splash', dest='printlogo', action='store_false',
             help='Prevent printing the greeting logo.', default=True)
    group.add_argument('-e', '--enable-interpreter', dest='interpreter',
             action='store_true', default=False,
             help='''Allow arbitrary single Python commands or blocks of Python code
             to be run. By default Python commands will not be run as a safeguard
             for your system. Make sure you trust the source of the ParmEd command
             before turning this option on.''')
    group = parser.add_argument_group('Error Handling', '''These options control how
             ParmEd handles various errors and warnings that appear occur during the
             course of Action execution''')
    group.add_argument('-s', '--strict', dest='strict', action='store_true',
             default=True, help='''Prevent scripts from running past unrecognized input and actions
             that end with an error. In interactive mode, actions with unrecognized inputs and
             failed actions prevent any changes from being made to the topology, but does not quit
             the interpreter. This is the default behavior.''')
    group.add_argument('-r', '--relaxed', dest='strict', action='store_false',
             help='''Scripts ignore unrecognized input and simply skip over failed
             actions, executing the rest of the script. Unrecognized input in the
             interactive interpreter emits a non-fatal warning.''')
    parser.add_argument('prmtop_cl', nargs='?', metavar='<prmtop>', default=None,
             help='Topology file to analyze.')
    parser.add_argument('script_cl', nargs='?', metavar='<script>', default=None,
             help='File with a series of ParmEd commands to execute.')

    opt = parser.parse_args()

    # If the user specified a prmtop and script in the 'old' way, append them to the
    # list constructed via the --parm and -i flags -- they come at the end
    if opt.script_cl is not None:
        opt.script.append(opt.script_cl)

    if opt.prmtop_cl is not None:
        opt.prmtop.append(opt.prmtop_cl)

    # Load the splash screen
    if opt.printlogo:
        splash = Logo()
        print(splash)

    # Set our warning filter
    if not opt.strict:
        warnings.filterwarnings('always', category=SeriousParmWarning)

    # Set our overwrite preferences
    Action.overwrite = opt.overwrite

    amber_prmtop = ParmList()
    for i, parm in enumerate(opt.prmtop):
        if i < len(opt.inpcrd):
            amber_prmtop.add_parm(parm)
            amber_prmtop.parm.load_rst7(opt.inpcrd[i])
            print(f'Loaded Amber topology file {parm} with coordinates from {opt.inpcrd[i]}\n')
        else:
            amber_prmtop.add_parm(parm)
            print(f'Loaded Amber topology file {parm}')

    if len(opt.script) > 0:
        # Read from the list of scripts
        print(opt.script)
        # Make sure that all scripts exist, quitting if we are strict and not all
        # scripts exist. Don't do anything until we know that all scripts exist.
        for script in opt.script:
            if not os.path.exists(script):
                warnings.warn(f'Script file {script} cannot be found.', SeriousParmWarning)

        # We have already pre-screened the scripts.
        for script in opt.script:
            if not os.path.exists(script):
                continue
            print(f'Reading actions from {script}\n')
            parmed_commands = ParmedCmd(amber_prmtop, stdin=open(script, 'r'))
            parmed_commands.use_rawinput = 0
            parmed_commands.interpreter = opt.interpreter
            parmed_commands.prompt = ''
            # Loop through all of the commands
            try:
                parmed_commands.cmdloop()
            except InterpreterError as err:
                sys.exit(f'{err.__class__.__name__}: {err}')
            except ParmedError:
                # This has already been caught and printed. If it was re-raised,
                # then that means we wanted to exit
                sys.exit(1)

    else:
        close_log_file = False
        parmed_commands = ParmedCmd(amber_prmtop)
        parmed_commands.intro = "Reading input from STDIN..."
        parmed_commands.interpreter = opt.interpreter
        parmed_commands.prompt = opt.prompt.strip() + ' '
        # Set the log file and logging, but only if interactive
        if os.isatty(sys.stdin.fileno()):
            if os.path.exists(opt.logfile) and readline is not None:
                # Load the logfile as a history, and get rid of any timestamps
                # from the history
                try:
                    f = open(opt.logfile, 'r')
                except IOError:
                    pass # Do not have permission or something
                else:
                    for line in f:
                        if line.startswith('# Log started on'):
                            continue
                        readline.add_history(line.rstrip())
                    f.close()
            try:
                logfile = open(opt.logfile, 'a')
            except IOError:
                print("Could not open logfile. Skip logging", file=sys.stderr)
                close_log_file = False
            else:
                now = datetime.datetime.now()
                logfile.write(f'# Log started on {now.month:02d}/{now.day:02d}/{now.year:d} '
                              f'[mm/dd/yyyy] at {now.hour:02d}:{now.minute:02d}:{now.second:02d}\n')
                if len(sys.argv) > 1:
                    logfile.write(f"# Command line arguments: {' '.join(sys.argv[1:])}\n")
                if len(amber_prmtop) > 0:
                    logfile.write(f"# Loaded topologies: {', '.join(parm.name for parm in amber_prmtop)}")
                parmed_commands.setlog(logfile)
                close_log_file = True
        # Loop through all of the commands
        try:
            try:
                parmed_commands.cmdloop()
            except InterpreterError as err:
                sys.exit(f"{err.__class__.__name__}: {err}")
            except ParmedError:
                # This has already been caught and printed. If it was re-raised,
                # then that means we wanted to exit
                sys.exit(1)
        finally:
            if close_log_file:
                logfile.close()

    print('Done!')



"""
Tests the ParmList class
"""
import os
from parmed import Structure, load_file
from parmed.amber import AmberParm, ChamberParm, AmoebaParm
from parmed.charmm import CharmmPsfFile
from parmed.tools import ParmList, exceptions as exc
import unittest
from utils import get_fn

class TestParmList(unittest.TestCase):

    def test_add_amber_parm(self):
        """ Test adding AmberParm-derived instances to ParmList """
        parms = ParmList()
        amber = AmberParm(get_fn('tip4p.parm7'))
        chamber = ChamberParm(get_fn('ala_ala_ala.parm7'))
        amoeba = AmoebaParm(get_fn('nma.parm7'))
        parms.add_parm(amber)
        parms.add_parm(chamber)
        parms.add_parm(amoeba)
        self.assertEqual(len(parms), 3)
        # Test int indexing
        self.assertIs(parms[0], amber)
        self.assertIs(parms[1], chamber)
        self.assertIs(parms[2], amoeba)
        self.assertRaises(IndexError, lambda: parms[3])
        # Test name indexing
        self.assertIs(parms[get_fn('tip4p.parm7')], amber)
        self.assertIs(parms[get_fn('ala_ala_ala.parm7')], chamber)
        self.assertIs(parms[get_fn('nma.parm7')], amoeba)
        self.assertRaises(IndexError, lambda: parms['noparm'])
        # Check that the last added parm is active
        self.assertIs(parms.parm, amoeba)
        # Set a new active topology
        parms.set_new_active(0)
        self.assertIs(parms.parm, amber)
        parms.set_new_active(get_fn('ala_ala_ala.parm7'))
        self.assertIs(parms.parm, chamber)
        self.assertRaises(IndexError, lambda: parms.set_new_active(3))
        self.assertIs(parms.parm, chamber)
        self.assertRaises(exc.DuplicateParm, lambda:
                parms.add_parm(get_fn('ala_ala_ala.parm7')))
        self.assertRaises(exc.DuplicateParm, lambda:
                parms.add_parm(load_file(get_fn('ala_ala_ala.parm7'))))
        self.assertRaises(exc.ParmError, lambda:
                parms.add_parm(os.path.join(get_fn('pptest1'), 'pptest1.h')))
        self.assertRaises(exc.ParmError, lambda:
                parms.add_parm(get_fn('amino12.lib')))
        self.assertRaises(exc.ParmError, lambda: parms.add_parm(object()))

    def test_add_amber_parm_names(self):
        """ Test adding Amber prmtop file names to ParmList """
        parms = ParmList()
        parms.add_parm(get_fn('tip4p.parm7'))
        parms.add_parm(get_fn('ala_ala_ala.parm7'))
        parms.add_parm(get_fn('nma.parm7'))
        amber, chamber, amoeba = parms
        self.assertEqual(len(parms), 3)
        # Check that they are the correct type
        self.assertIs(type(parms[0]), AmberParm)
        self.assertIs(type(parms[1]), ChamberParm)
        self.assertIs(type(parms[2]), AmoebaParm)
        self.assertRaises(IndexError, lambda: parms[3])
        # Test name indexing
        self.assertIs(parms[get_fn('tip4p.parm7')], amber)
        self.assertIs(parms[get_fn('ala_ala_ala.parm7')], chamber)
        self.assertIs(parms[get_fn('nma.parm7')], amoeba)
        self.assertRaises(IndexError, lambda: parms['noparm'])
        # Check that the last added parm is active
        self.assertIs(parms.parm, amoeba)
        # Set a new active topology
        parms.set_new_active(0)
        self.assertIs(parms.parm, amber)
        parms.set_new_active(get_fn('ala_ala_ala.parm7'))
        self.assertIs(parms.parm, chamber)
        self.assertRaises(IndexError, lambda: parms.set_new_active(3))
        self.assertIs(parms.parm, chamber)

    def test_add_charmm_psf(self):
        """ Test adding CHARMM PSF file to ParmList """
        parms = ParmList()
        ala3 = CharmmPsfFile(get_fn('ala3_solv.psf'))
        aaa = CharmmPsfFile(get_fn('ala_ala_ala.psf'))
        xaaa = CharmmPsfFile(get_fn('ala_ala_ala.psf.xplor'))
        aaaa = CharmmPsfFile(get_fn('ala_ala_ala_autopsf.psf'))
        parms.add_parm(ala3)
        parms.add_parm(aaa)
        parms.add_parm(xaaa)
        parms.add_parm(aaaa)
        # Check indexing
        self.assertIs(parms[0], ala3)
        self.assertIs(parms[1], aaa)
        self.assertIs(parms[2], xaaa)
        self.assertIs(parms[3], aaaa)
        # Test name indexing
        self.assertIs(parms[get_fn('ala3_solv.psf')], ala3)
        self.assertIs(parms[get_fn('ala_ala_ala.psf')], aaa)
        self.assertIs(parms[get_fn('ala_ala_ala.psf.xplor')], xaaa)
        self.assertIs(parms[get_fn('ala_ala_ala_autopsf.psf')], aaaa)
        # Check that the last parm added is active
        self.assertIs(parms.parm, parms[-1])
        # Set a new active topology
        parms.set_new_active(0)
        self.assertIs(parms.parm, ala3)
        parms.set_new_active(get_fn('ala_ala_ala.psf'))
        self.assertIs(parms.parm, aaa)


    def test_add_charmm_psf_names(self):
        """ Test adding CHARMM PSF file names to ParmList """
        parms = ParmList()
        parms.add_parm(get_fn('ala3_solv.psf'))
        parms.add_parm(get_fn('ala_ala_ala.psf'))
        parms.add_parm(get_fn('ala_ala_ala.psf.xplor'))
        parms.add_parm(get_fn('ala_ala_ala_autopsf.psf'))
        ala3, aaa, xaaa, aaaa = parms
        # Check proper instantiation
        for parm in parms:
            self.assertIsInstance(parm, CharmmPsfFile)
        # Check indexing
        self.assertIs(parms[0], ala3)
        self.assertIs(parms[1], aaa)
        self.assertIs(parms[2], xaaa)
        self.assertIs(parms[3], aaaa)
        # Test name indexing
        self.assertIs(parms[get_fn('ala3_solv.psf')], ala3)
        self.assertIs(parms[get_fn('ala_ala_ala.psf')], aaa)
        self.assertIs(parms[get_fn('ala_ala_ala.psf.xplor')], xaaa)
        self.assertIs(parms[get_fn('ala_ala_ala_autopsf.psf')], aaaa)
        # Check that the last parm added is active
        self.assertIs(parms.parm, parms[-1])
        # Set a new active topology
        parms.set_new_active(0)
        self.assertIs(parms.parm, ala3)
        parms.set_new_active(get_fn('ala_ala_ala.psf'))
        self.assertIs(parms.parm, aaa)

    def test_add_mol2(self):
        """ Test adding Mol2 file to ParmList """
        parms = ParmList()
        parms.add_parm(get_fn('tripos9.mol2'))
        self.assertEqual(len(parms), 1)
        self.assertIsInstance(parms[0], Structure)
        self.assertIn(0, parms)
        self.assertNotIn(1, parms)



"""
Contains all of the class objects for the Tinker topology
"""
from collections import OrderedDict

class Atom(object):
    """ An atom in the system """

    def __init__(self, symbol, type_, class_, 
                 atomic_number, mass, valence, desc):
        self.symbol = str(symbol).strip()
        self.type = int(type_)
        self.class_ = int(class_)
        self.atomic_number = int(atomic_number)
        self.mass = float(mass)
        self.valence = int(valence)
        self.desc = str(desc).strip()

    def add_vdw(self, radius, epsilon, radius14, epsilon14, reduction):
        """ Adds van der Waals terms to the atom """
        self.radius = float(radius)
        self.epsilon = float(epsilon)
        # These values might not exist
        if not str(radius14).strip():
            self.radius14 = None
        else:
            self.radius14 = float(radius14)
        if not str(epsilon14).strip():
            self.epsilon14 = None
        else:
            self.epsilon14 = float(epsilon14)
        if not str(reduction).strip():
            self.reduction = None
        else:
            self.reduction = float(reduction)

    def __repr__(self):
        return f'<Atom {self.symbol} [#{self.atomic_number}]; Mass={self.mass:.2f}; "{self.desc}">'

    def __str__(self):
        retstr =  (f'Atom {self.symbol} [#{self.atomic_number}] {self.desc}\n'
                   f'   Type:        {self.type}\n'
                   f'   Class:       {self.class_}\n'
                   f'   Mass:        {self.mass:.2f}\n'
                   f'   Valence:     {self.valence}')

        if hasattr(self, 'radius'):
            retstr += (f'\n   vdW Rad:     {self.radius:.2f}\n'
                       f'   vdW Eps:     {self.epsilon:.2f}')
            if self.radius14 is not None or self.epsilon14 is not None:
                retstr += f'\n   vdW 1-4 Rad: {self.radius14:.2f}\n   vdW 1-4 Eps: {self.epsilon14:.2f}'

        if self.reduction is not None:
            retstr += f'\n   vdW Reduct:  {self.reduction:.2f}'

        return retstr

class _ParamTypeList(list):
    """
    A list for all types of parameters. This should be subclassed for each
    parameter type that's necessary. It performs type checking as well as other
    tasks.
    """

    typeclass = None

    def __init__(self):
        super().__init__()

    def add(self, *args, **kwargs):
        self.append(self.typeclass(*args, **kwargs))

    def append(self, thing):
        if not isinstance(thing, self.typeclass):
            raise TypeError(f'Can only append "{self.typeclass.__name__}" objects to {self.__class__.__name__}')
        super().append(thing)

    def extend(self, things):
        for thing in things:
            self.append(thing)

class AtomList(_ParamTypeList):
    """ List for Atom instances """
    typeclass = Atom

class BondStretch(object):
    """ A bond-stretching term """
    def __init__(self, atom1, atom2, k, req):
        self.atom1 = atom1
        self.atom2 = atom2
        self.k = float(k)
        self.req = float(req)

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.atom1.symbol}-{self.atom2.symbol}]; k={self.k:.2f}; Req={self.req:.2f}>'

    def __str__(self):
        return (f'{self.__class__.name__} {repr(self.atom1)} --- {repr(self.atom2)}\n'
                f'     K = {self.k:.4f}\n'
                f'   Req = {self.req:.4f}')

class BondList(_ParamTypeList):
    typeclass = BondStretch

class AngleBend(object):
    """ An angle-bending term """
    def __init__(self, atom1, atom2, atom3, k, theteq, fold, type_):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.k = float(k)
        self.theteq = float(theteq)
        if str(fold).strip():
            self.fold = float(fold)
        else:
            self.fold = None
        self.type = type_.strip() or 'Harmonic'

    def __repr__(self):
        retstr = (
            f'<AngleBend [{self.atom1.symbol}-{self.atom2.symbol}-{self.atom3.symbol}]; '
            f'k={self.k:.2f}; Theta_eq={self.theteq:.2f}'
        )
        if self.fold is not None:
            retstr += f'; fold={self.fold:.1f}'
        return retstr + f'; {self.type}>'

    def __str__(self):
        retstr = (f'AngleBend {repr(self.atom1)} --- {repr(self.atom2)} --- {repr(self.atom3)}\n'
                  f'     K = {self.k:.4f}\n'
                  f' Theta = {self.theteq:.4f}\n')
        if self.fold is not None:
            retstr += f'  Fold = {self.fold:.1f}\n'
        return retstr + f'  Type = {self.type}'

class AngleList(_ParamTypeList):
    typeclass = AngleBend

class StretchBend(object):
    """ A stretch-bending term """
    def __init__(self, atom1, atom2, atom3, k, theteq, r1eq, r2eq):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.k = float(k)
        self.theteq = float(theteq)
        self.r1eq = float(r1eq)
        self.r2eq = float(r2eq)

    def __repr__(self):
        return (
            f'<StretchBend [{self.atom1.symbol}-{self.atom2.symbol}-{self.atom3.symbol}]; '
            f'k={self.k:.2f}; Theta_eq={self.theteq:.2f}; R1eq={self.r1eq:.2f}; R2eq={self.r2eq:.2f}>'
        )

    def __str__(self):
        return (f'StretchBend {repr(self.atom1)} --- {repr(self.atom2)} --- {repr(self.atom3)}\n'
                f'     K = {self.k:.4f}\n'
                f' Theta = {self.theteq:.4f}\n'
                f'  R1eq = {self.r1eq:.4f}\n'
                f'  R2eq = {self.r2eq:.4f}')

class StretchBendList(_ParamTypeList):
    typeclass = StretchBend

class UreyBradley(BondStretch):
    """ A urey-bradley term -- functionally identical to a bond-stretch """

class UreyBradleyList(_ParamTypeList):
    typeclass = UreyBradley

class OutOfPlaneBend:
    """ An out-of-plane bending term """
    def __init__(self, atom1, atom2, atom3, atom4, k):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.atom4 = atom4
        self.k = float(k)

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.atom1.symbol}-{self.atom2.symbol}-{self.atom3.symbol}-{self.atom4.symbol}]; k={self.k:.2f}>'

    def __str__(self):
        return (f'{self.__class__.__name__} {repr(self.atom1)} --- {repr(self.atom2)} --- {repr(self.atom3)} --- {repr(self.atom4)}\n'
                f'   K = {self.k:.2f}')

class OutOfPlaneBendList(_ParamTypeList):
    typeclass = OutOfPlaneBend

class OutOfPlaneDist(OutOfPlaneBend):
    """ An out-of-plane distance (functionally equivalent to OOP Bend """

class OutOfPlaneDistList(_ParamTypeList):
    typeclass = OutOfPlaneDist

class TorsionAngle(object):
    """ Torsional Angle parameter """
    def __init__(self, atom1, atom2, atom3, atom4, args):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.atom4 = atom4
        # We don't know how many terms to expect, but it will be a multiple of 3
        if len(args) % 3 != 0:
            raise TypeError('TorsionAngle expects an equal number of '
                            'Amplitudes, phases, and periodicities.')
        nterms = len(args) // 3
        self.amplitude = tuple([float(args[3*i]) for i in range(nterms)])
        self.phase = tuple([float(args[3*i+1]) for i in range(nterms)])
        self.periodicity = tuple([int(args[3*i+2]) for i in range(nterms)])
        if (len(self.amplitude) != len(self.phase) or
                len(self.amplitude) != len(self.periodicity)):
            raise RuntimeError('BUGBUG!! Inconsistent # of terms in torsion')

    def __repr__(self):
        if len(self.amplitude) == 0:
            return f'<TorsionAngle [{self.atom1.symbol}-{self.atom1.symbol}-{self.atom1.symbol}-{self.atom1.symbol}]; No Terms>'
        return (f'<TorsionAngle [{self.atom1.symbol}-{self.atom1.symbol}-{self.atom1.symbol}-{self.atom1.symbol}]; '
                f'Ampl={self.amplitude}; Phase={self.phase}; Per={self.periodicity}>')

    def __str__(self):
        if len(self.amplitude) == 0:
            return (f'TorsionAngle {repr(self.atom1)} --- {repr(self.atom2)} --- {repr(self.atom3)} --- {repr(self.atom4)}\n'
                    '   No Terms')
        retstr = f'TorsionAngle {repr(self.atom1)} --- {repr(self.atom2)} --- {repr(self.atom3)} --- {repr(self.atom4)}'
        seq = range(len(self.amplitude))
        for i, amp, phase, per in enumerate(zip(seq, self.amplitude, self.phase, self.periodicity)):
            retstr += (f'\n   Term {i + 1}\n'
                       f'   Amplitude = {amp:.4f}\n'
                       f'       Phase = {phase:.4f}\n'
                       f' Periodicity = {per:.4f}')
        return retstr

class TorsionAngleList(_ParamTypeList):
    typeclass = TorsionAngle

class PiTorsion(object):
    """ A Pi-Orbital Torsion parameter """
    def __init__(self, atom1, atom2, amplitude):
        self.atom1, self.atom2 = atom1, atom2
        self.amplitude = float(amplitude)

    def __repr__(self):
        return f'<PiTorsion [{self.atom1.symbol}-{self.atom2.symbol}]; Ampl={self.amplitude:.2f}>'

    def __str__(self):
        return f'PiTorsion {repr(self.atom1)} --- {repr(self.atom1)}\n   Amplitude = {self.amplitude:.4f}'

class PiTorsionList(_ParamTypeList):
    typeclass = PiTorsion

class TorsionTorsion(object):
    """ A coupled-torsion parameter (like CMAP) """
    def __init__(self, atom1, atom2, atom3, atom4, atom5, spline1, spline2):
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.atom4 = atom4
        self.atom5 = atom5
        self.spline1 = int(spline1)
        self.spline2 = int(spline2)

    def __repr__(self):
        return (
            f'<TorsionTorsion [{self.atom1.symbol}-{self.atom2.symbol}-{self.atom3.symbol}-{self.atom4.symbol}-{self.atom5.symbol}]; '
            f'Spline=({self.spline1},{self.spline2})>'
        )

    def __str__(self):
        return (f'TorsionTorsion {repr(self.atom1)} --- {repr(self.atom2)} --- {repr(self.atom3)} ---\n'
                f'                      {repr(self.atom4)} --- {repr(self.atom5)}\n'
                f'   Spline Grid = ({self.spline1} x {self.spline2})')

class TorsionTorsionList(_ParamTypeList):
    typeclass = TorsionTorsion

class TorsionTorsionGrid(OrderedDict):
    """
    The interpolation grid of a coupled-torsion correction map. Every unique
    grid is cached and if a duplicate grid is instantiated, a reference to the
    original grid is returned. As a result, all unique TorsionTorsionGrid
    instances are singletons and should be compared for equality with "is"
    """
    # Because these grids are space-consuming, we only hold onto unique grids
    _typelist = list()

    @classmethod
    def new(cls, data):
        inst = cls()
        for d in data:
            inst[tuple(d[:2])] = d[2]
        # Potentially expensive comparison of all grids.
        for ttg in TorsionTorsionGrid._typelist:
            if ttg == inst:
                return ttg
        TorsionTorsionGrid._typelist.append(inst)
        return inst
   
    def __eq__(self, other):
        if self.keys() != other.keys(): return False
        for key in self:
            if abs(self[key] - other[key]) > 1e-8: return False
        return True
   
    def __ne__(self, other):
        return not self == other

    # No comparisons are implemented
    def __gt__(self, other):
        raise NotImplemented('TorsionTorsionGrid instances are not well-ordered')
    __lt__ = __ge__ = __le__ = __gt__

class AtomicMultipole(object):
    """ Atomic multipole parameter """
    def __init__(self, atom, frame, definition, moments):
        self.atom = atom
        self.frame = [int(i) for i in frame if str(i).strip()]
        self.definition = definition.strip()
        self.moment = [float(x) for x in moments]

    def __repr__(self):
        return f'<AtomicMultipole [{self.atom.symbol}] {self.definition}; Frame={self.frame}; Moments={self.moment}>'

    def __str__(self):
        return (f'AtomicMultipole {repr(self.atom)} "{self.definition}"\n'
                f'     Frame = {self.frame}\n'
                f'   Moments = {self.moment[0]:8.5f}\n'
                f'             {self.moment[1]:8.5f} {self.moment[2]:8.5f} {self.moment[3]:8.5f}\n'
                f'             {self.moment[4]:8.5f}\n'
                f'             {self.moment[5]:8.5f} {self.moment[6]:8.5f}\n'
                f'             {self.moment[7]:8.5f} {self.moment[8]:8.5f} {self.moment[9]:8.5f}')

class AtomicMultipoleList(_ParamTypeList):
    typeclass = AtomicMultipole

class DipolePolarizability(object):
    """ A dipole polarizability parameter """
    def __init__(self, atom, alpha, group):
        self.atom = atom
        self.alpha = float(alpha)
        self.group = [int(g) for g in group]

    def __repr__(self):
        return f'<DipolePolarizability [{self.atom.symbol}]; Alpha={self.alpha:.2f}; Group={self.group}>'

    def __str__(self):
        return (f'DipolePolarizability {repr(self.atom)}\n'
                f'    Alpha = {self.alpha:.4f}\n'
                f'    Group = {self.group}')

class DipolePolarizabilityList(_ParamTypeList):
    typeclass = DipolePolarizability



#!/usr/bin/env python
from __future__ import division, print_function

import sys

# OpenMM Imports
import openmm as mm
import openmm.app as app

# ParmEd Imports
from parmed import load_file, unit as u
from parmed.openmm import StateDataReporter, NetCDFReporter

# Load the Amber files
print('Loading AMBER files...')
ala2_solv = load_file('ala2_solv.parm7', 'ala2_solv.rst7')

# Create the OpenMM system
print('Creating OpenMM System')
system = ala2_solv.createSystem(nonbondedMethod=app.PME,
                                nonbondedCutoff=8.0*u.angstroms,
                                constraints=app.HBonds,
)

# Create the integrator to do Langevin dynamics
integrator = mm.LangevinIntegrator(
                        300*u.kelvin,       # Temperature of heat bath
                        1.0/u.picoseconds,  # Friction coefficient
                        2.0*u.femtoseconds, # Time step
)

# Define the platform to use; CUDA, OpenCL, CPU, or Reference. Or do not specify
# the platform to use the default (fastest) platform
platform = mm.Platform.getPlatformByName('CUDA')
prop = dict(CudaPrecision='mixed') # Use mixed single/double precision

# Create the Simulation object
sim = app.Simulation(ala2_solv.topology, system, integrator, platform, prop)

# Set the particle positions
sim.context.setPositions(ala2_solv.positions)

# Minimize the energy
print('Minimizing energy')
sim.minimizeEnergy(maxIterations=500)

# Set up the reporters to report energies and coordinates every 100 steps
sim.reporters.append(
        StateDataReporter(sys.stdout, 100, step=True, potentialEnergy=True,
                          kineticEnergy=True, temperature=True, volume=True,
                          density=True)
)
sim.reporters.append(NetCDFReporter('ala2_solv.nc', 100, crds=True))

# Run dynamics
print('Running dynamics')
sim.step(10000)



import sys
import parmed as pmd
import unittest
import numpy as np

from utils import get_fn

try:
    import rdkit
    has_rdkit = True
except ImportError:
    has_rdkit = False

@unittest.skipUnless(has_rdkit, "Only test load_rdkit module on Linux")
class TestRDKit(unittest.TestCase):
    """ Tests loading of an rdkit Mol object """

    def test_load_rdkit_mol(self):
        """ test load rdkit from Mol """
        from rdkit import Chem
        m1 = Chem.MolFromSmiles('C1=CC=CN=C1')
        parm = pmd.load_rdkit(m1)
        self.assertEqual([atom.name for atom in parm.atoms], ['C1', 'C2', 'C3', 'C4', 'N1', 'C5']) 
        self.assertEqual(parm.residues[0].name, 'UNL')

    def test_load_smiles(self):
        """ test load rdkit from smiles string """
        smiles = 'C1=CC=CN=C1'

        # coordinates = False
        parm = pmd.rdkit.from_smiles(smiles, coordinates=False, hydrogens=False)
        self.assertEqual([atom.name for atom in parm.atoms], ['C1', 'C2', 'C3', 'C4', 'N1', 'C5']) 
        self.assertEqual(parm.residues[0].name, 'UNL')
        self.assertIs(parm.coordinates, None)
        self.assertIs(parm.get_coordinates(), None)

        # coordinates = True (default)
        parm = pmd.rdkit.from_smiles(smiles, coordinates=True, hydrogens=False)
        self.assertFalse((parm.coordinates[0] == 0).all())

    def test_load_smiles_explicit_hydrogen(self):
        """ test adding explict hydrogens from smiles string"""
        smiles = "CC"

        num_atoms = 8 # CH3-CH3
        parm = pmd.rdkit.from_smiles(smiles=smiles, coordinates=False, hydrogens=True)
        self.assertEqual(len(parm.atoms), num_atoms)
        self.assertEqual([atom.name for atom in parm.atoms], ['C1', 'C2', 'H1', 'H2', 'H3', 'H4', 'H5', 'H6'])

    def test_to_mol(self):
        """ Test converting a Structure to an RDKit Mol object """
        from rdkit.Chem import Mol

        structure = pmd.load_file(get_fn("4lzt.pdb"))
        mol = structure.rdkit_mol

        self.assertIsInstance(mol, Mol)
        self.assertEqual(mol.GetNumAtoms(), len(structure.atoms))
        self.assertEqual(mol.GetNumBonds(), len(structure.bonds))



"""
List of all pointers and constants used in the Amber topology file.

Can be used like:
   from parmed.constants import *
"""
from enum import IntEnum

from math import pi as _pi, sqrt as _sqrt, log10 as _log10, acos as _acos

__all__ = ['AMBER_ELECTROSTATIC', 'AMBER_POINTERS', 'PrmtopPointers',
           'RAD_TO_DEG', 'DEG_TO_RAD', 'TRUNCATED_OCTAHEDRON_ANGLE']

AMBER_ELECTROSTATIC = 18.2223
CHARMM_ELECTROSTATIC = _sqrt(332.0716)

AMBER_POINTERS = """
NATOM  : total number of atoms
NTYPES : total number of distinct atom types
NBONH  : number of bonds containing hydrogen
MBONA  : number of bonds not containing hydrogen
NTHETH : number of angles containing hydrogen
MTHETA : number of angles not containing hydrogen
NPHIH  : number of dihedrals containing hydrogen
MPHIA  : number of dihedrals not containing hydrogen
NHPARM : currently not used
NPARM  : currently not used
NEXT   : number of excluded atoms
NRES   : number of residues
NBONA  : MBONA + number of constraint bonds
NTHETA : MTHETA + number of constraint angles
NPHIA  : MPHIA + number of constraint dihedrals
NUMBND : number of unique bond types
NUMANG : number of unique angle types
NPTRA  : number of unique dihedral types
NATYP  : number of atom types in parameter file, see SOLTY below
NPHB   : number of distinct 10-12 hydrogen bond pair types
IFPERT : set to 1 if perturbation info is to be read in
NBPER  : number of bonds to be perturbed
NGPER  : number of angles to be perturbed
NDPER  : number of dihedrals to be perturbed
MBPER  : number of bonds with atoms completely in perturbed group
MGPER  : number of angles with atoms completely in perturbed group
MDPER  : number of dihedrals with atoms completely in perturbed groups
IFBOX  : set to 1 if standard periodic box, 2 when truncated octahedral
NMXRS  : number of atoms in the largest residue
IFCAP  : set to 1 if the CAP option from edit was specified
NUMEXTRA: number of extra points
NCOPY  : Number of copies for advanded simulations
"""
# These global variables provide a more natural way of accessing
# the various pointers.  Most useful if they're loaded into the
# top-level namespace.
class PrmtopPointers(IntEnum):
    NATOM = 0
    NTYPES = 1
    NBONH = 2
    MBONA = 3
    NTHETH = 4
    MTHETA = 5
    NPHIH = 6
    MPHIA = 7
    NHPARM = 8
    NPARM = 9
    NEXT = 10
    NRES = 11
    NBONA = 12
    NTHETA = 13
    NPHIA = 14
    NUMBND = 15
    NUMANG = 16
    NPTRA = 17
    NATYP = 18
    NPHB = 19
    IFPERT = 20
    NBPER = 21
    NGPER = 22
    NDPER = 23
    MBPER = 24
    MGPER = 25
    MDPER = 26
    IFBOX = 27
    NMXRS = 28
    IFCAP = 29
    NUMEXTRA = 30
    NCOPY = 31
    NNB = NEXT

RAD_TO_DEG = 180.0 / _pi
DEG_TO_RAD = _pi / 180.0
TRUNCATED_OCTAHEDRON_ANGLE = _acos(-1/3) * 180 / _pi

# For use in floating point comparisons
TINY = 1.0e-8
SMALL = 1.0e-4
TINY_DIGITS = int(_log10(TINY) + 0.5)
SMALL_DIGITS = int(_log10(SMALL) + 0.5)

# For I/O
DEFAULT_ENCODING = 'UTF-8'



"""
Contains unittests for running OpenMM calculations using the Amber file parsers
"""
from collections import defaultdict
from copy import copy
from math import sqrt
import os
from parmed import Structure, topologyobjects, load_file
from parmed.exceptions import ParameterError
from parmed.amber import AmberParm, ChamberParm, Rst7
from parmed.openmm import load_topology, energy_decomposition_system
import parmed.unit as u
import parmed.tools as PT
import sys
import unittest
from utils import get_fn, CPU, mm, app, has_openmm, TestCaseRelative, QuantityTestCase

# OpenMM NonbondedForce methods are enumerated values. From NonbondedForce.h,
# they are:
#   0 - NoCutoff
#   1 - CutoffNonPeriodic
#   2 - CutoffPeriodic
#   3 - Ewald
#   4 - PME

def energy_decomposition(parm, context):
    from parmed.openmm.utils import energy_decomposition
    ret = defaultdict(float)
    for key, val in energy_decomposition(parm, context).items():
        ret[key] = val
    return ret

@unittest.skipUnless(has_openmm, "Cannot test without OpenMM")
class TestAmberParm(TestCaseRelative, QuantityTestCase):

    def test_gas_energy_conf_1(self):
        """ Compare Amber and OpenMM gas phase energies and forces (topology 1) """
        parm = AmberParm(
            get_fn('hydrogen-peroxide_T0_1.prmtop'), get_fn('hydrogen-peroxide_T0_1.rst7')
        )
        self.assertEqual(parm.combining_rule, 'lorentz')
        system = parm.createSystem() # Default, no cutoff
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(parm.positions)
        energies = energy_decomposition(parm, sim.context)
#BOND    =        0.0319  ANGLE   =        2.1690  DIHED      =        0.2218
#VDWAALS =        0.0000  EEL     =        0.0000  HBOND      =        0.0000
#1-4 VDW =        0.0000  1-4 EEL =        0.0000  RESTRAINT  =        0.0000
        # Compare OpenMM energies with the Amber energies (above)
        self.assertAlmostEqual(energies['bond'], 0.0319, places=4)
        self.assertAlmostEqual(energies['angle'], 2.1690, places=4)
        # Cannot compare torsional potentials because Amber has a 0-periodicity term that shifts
        # the energy by a constant (does not impact forces), and OpenMM has stopped supporting those
        # terms.
        # self.assertAlmostEqual(energies['dihedral'], 0.2218, places=4)
        self.assertRelativeEqual(energies['nonbonded'], 0.0, places=3)

        # Now test the forces to make sure that they are computed correctly in
        # the presence of extra points
        pstate = sim.context.getState(getForces=True)
        pf = pstate.getForces().value_in_unit(u.kilojoule_per_mole/u.nanometer)
# gmx forces:
#     f[    0]={ 5.37936e+02, -5.90363e+02, -6.71315e+01}
#     f[    1]={-5.37936e+02,  5.90363e+02, -6.71315e+01}
#     f[    2]={ 9.92309e+01,  5.90363e+02,  6.71315e+01}
#     f[    3]={-9.92309e+01, -5.90363e+02,  6.71315e+01}
        gf = [  ( 5.37936e+02, -5.90363e+02, -6.71315e+01),
                (-5.37936e+02,  5.90363e+02, -6.71315e+01),
                ( 9.92309e+01,  5.90363e+02,  6.71315e+01),
                (-9.92309e+01, -5.90363e+02,  6.71315e+01)]
        for p, s in zip(pf, gf):
            for x1, x2 in zip(p, s):
                self.assertAlmostEqual(x1, x2, places=3)

    def test_gas_energy_conf_2(self):
        """ Compare Amber and OpenMM gas phase energies and forces (topology 2) """
        parm = AmberParm(
            get_fn('hydrogen-peroxide_T0_2.prmtop'), get_fn('hydrogen-peroxide_T0_2.rst7')
        )
        self.assertEqual(parm.combining_rule, 'lorentz')
        system = parm.createSystem() # Default, no cutoff
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(parm.positions)
        energies = energy_decomposition(parm, sim.context)
#BOND    =        0.0319  ANGLE   =        2.1690  DIHED      =       -2.7931
#VDWAALS =        0.0000  EEL     =        0.0000  HBOND      =        0.0000
#1-4 VDW =        0.0000  1-4 EEL =        0.0000  RESTRAINT  =        0.0000
        # Compare OpenMM energies with the Amber energies (above)
        self.assertAlmostEqual(energies['bond'], 0.0319, places=4)
        self.assertAlmostEqual(energies['angle'], 2.1690, places=4)
        # Cannot compare torsional potentials because Amber has a 0-periodicity term that shifts
        # the energy by a constant (does not impact forces), and OpenMM has stopped supporting those
        # terms.
        # self.assertAlmostEqual(energies['dihedral'],-2.7931, places=4)
        self.assertRelativeEqual(energies['nonbonded'], 0.0, places=3)

        # Now test the forces to make sure that they are computed correctly in
        # the presence of extra points
        pstate = sim.context.getState(getForces=True)
        pf = pstate.getForces().value_in_unit(u.kilojoule_per_mole/u.nanometer)
# gmx forces:
#     f[    0]={ 3.14357e+02, -5.90363e+02,  2.11410e+02}
#     f[    1]={-3.14357e+02,  5.90363e+02,  2.11410e+02}
#     f[    2]={ 2.70641e+02,  5.90363e+02, -2.11410e+02}
#     f[    3]={-2.70641e+02, -5.90363e+02, -2.11410e+02}
        gf = [  ( 3.14357e+02, -5.90363e+02,  2.11410e+02), 
                (-3.14357e+02,  5.90363e+02,  2.11410e+02), 
                ( 2.70641e+02,  5.90363e+02, -2.11410e+02), 
                (-2.70641e+02, -5.90363e+02, -2.11410e+02)] 
        for p, s in zip(pf, gf):
            for x1, x2 in zip(p, s):
                self.assertAlmostEqual(x1, x2, places=3)

    def test_gas_energy_conf_3(self):
        """ Compare Amber and OpenMM gas phase energies and forces (topology 3) """
        parm = AmberParm(get_fn('methanol_T0.prmtop'), get_fn('methanol_T0.rst7'))
        self.assertEqual(parm.combining_rule, 'geometric')
        system = parm.createSystem() # Default, no cutoff
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(parm.positions)
        energies = energy_decomposition(parm, sim.context)
#BOND    =        0.0224  ANGLE   =        0.0469  DIHED      =        0.0039
#VDWAALS =        0.0000  EEL     =        0.0000  HBOND      =        0.0000
#1-4 VDW =        0.0000  1-4 EEL =        3.3789  RESTRAINT  =        0.0000
        # Compare OpenMM energies with the Amber energies (above)
        self.assertAlmostEqual(energies['bond'], 0.0224, places=4)
        self.assertAlmostEqual(energies['angle'], 0.0469, places=4)
        self.assertAlmostEqual(energies['dihedral'], 0.0039, places=4)
        self.assertRelativeEqual(energies['nonbonded'], 3.3789, places=3)

        # Now test the forces to make sure that they are computed correctly in
        # the presence of extra points
        pstate = sim.context.getState(getForces=True)
        pf = pstate.getForces().value_in_unit(u.kilojoule_per_mole/u.nanometer)
# gmx forces:
#     f[    0]={ 2.19935e+01,  8.60143e+01,  1.77895e+02}
#     f[    1]={ 5.25347e+01,  3.46119e+01, -9.73738e+01}
#     f[    2]={ 1.76138e+01, -7.13778e+01, -1.08807e+00}
#     f[    3]={-3.74994e+01, -5.39602e+01, -6.75679e+01}
#     f[    4]={-8.35142e+01, -7.86616e+01, -1.60611e+02}
#     f[    5]={ 2.88716e+01,  8.33735e+01,  1.48746e+02}

        gf = [  ( 2.19935e+01,  8.60143e+01,  1.77895e+02),
                ( 5.25347e+01,  3.46119e+01, -9.73738e+01),
                ( 1.76138e+01, -7.13778e+01, -1.08807e+00),
                (-3.74994e+01, -5.39602e+01, -6.75679e+01),
                (-8.35142e+01, -7.86616e+01, -1.60611e+02),
                ( 2.88716e+01,  8.33735e+01,  1.48746e+02)]
        for p, s in zip(pf, gf):
            for x1, x2 in zip(p, s):
                self.assertAlmostEqual(x1, x2, places=3)


    def test_gas_energy_conf_4(self):
        """ Compare Amber and OpenMM gas phase energies and forces (topology 4) """
        parm = AmberParm(get_fn('ethanol_T0.prmtop'), get_fn('ethanol_T0.rst7'))
        self.assertEqual(parm.combining_rule, 'geometric')
        system = parm.createSystem() # Default, no cutoff
        integrator = mm.VerletIntegrator(1.0*u.femtoseconds)
        sim = app.Simulation(parm.topology, system, integrator, platform=CPU)
        sim.context.setPositions(parm.positions)
        energies = energy_decomposition(parm, sim.context)
#BOND    =        0.0239  ANGLE   =        0.0298  DIHED      =        0.0093
#VDWAALS =        0.0000  EEL     =        6.7526  HBOND      =        0.0000
#1-4 VDW =        0.0492  1-4 EEL =       -6.0430  RESTRAINT  =        0.0000
        # Compare OpenMM energies with the Amber energies (above)
        self.assertAlmostEqual(energies['bond'], 0.0239, places=4)
        self.assertAlmostEqual(energies['angle'], 0.0298, places=4)
        # Cannot compare torsional potentials because Amber has a 0-periodicity term that shifts
        # the energy by a constant (does not impact forces), and OpenMM has stopped supporting those
        # terms.
        # self.assertAlmostEqual(energies['dihedral'], 0.0093, places=4)
        self.assertRelativeEqual(energies['nonbonded'], 0.0000+6.7526+0.0492-6.0430, places=3)

        # Now test the forces to make sure that they are computed correctly in
        # the presence of extra points
        pstate = sim.context.getState(getForces=True)
        pf = pstate.getForces().value_in_unit(u.kilojoule_per_mole/u.nanometer)
# gmx forces:
#      f[    0]={ 1.73101e+02,  5.67250e+01, -4.02950e+01}
#      f[    1]={-8.13704e+00, -1.79612e+01,  9.88537e+01}
#      f[    2]={-2.83120e+01,  6.23352e+00, -5.47393e+00}
#      f[    3]={-1.03312e+01, -1.00966e+01, -5.12129e+00}
#      f[    4]={-1.69636e+02,  3.34850e+01, -1.73612e+02}
#      f[    5]={ 4.19932e+00, -2.58283e+00,  1.29999e+02}
#      f[    6]={ 3.02865e+01, -6.68331e+00, -2.99153e+01}
#      f[    7]={ 1.00113e+02, -5.22480e+01,  4.80526e+01}
#      f[    8]={-9.12828e+01, -6.87157e+00, -2.24877e+01}

        gf = [  ( 1.73101e+02,  5.67250e+01, -4.02950e+01),
                (-8.13704e+00, -1.79612e+01,  9.88537e+01),
                (-2.83120e+01,  6.23352e+00, -5.47393e+00),
                (-1.03312e+01, -1.00966e+01, -5.12129e+00),
                (-1.69636e+02,  3.34850e+01, -1.73612e+02),
                ( 4.19932e+00, -2.58283e+00,  1.29999e+02),
                ( 3.02865e+01, -6.68331e+00, -2.99153e+01),
                ( 1.00113e+02, -5.22480e+01,  4.80526e+01),
                (-9.12828e+01, -6.87157e+00, -2.24877e+01)]
        for p, s in zip(pf, gf):
            for x1, x2 in zip(p, s):
                self.assertAlmostEqual(x1, x2, places=3)





"""
Unittests for serializing various objects in ParmEd
"""
from io import BytesIO
import numpy as np
import os
import parmed as pmd
import pickle
import random
from string import ascii_uppercase as uppercase
import unittest
import utils
from parmed.topologyobjects import QualitativeBondType
from utils import HAS_GROMACS

class TestParmedSerialization(unittest.TestCase):
    """ Tests ParmEd serialization """

    _equal_atoms = utils.equal_atoms

    def test_atom_serialization(self):
        """ Tests the serialization of Atom """
        atom = pmd.Atom(atomic_number=random.randint(1, 100),
                        name=random.choice(uppercase)+random.choice(uppercase),
                        type=random.choice(uppercase)+random.choice(uppercase),
                        charge=random.random()*2-1, mass=random.random()*30+1,
                        nb_idx=random.randint(1, 20),
                        solvent_radius=random.random()*2,
                        screen=random.random()*2, tree='M',
                        join=random.random()*2, irotat=random.random(),
                        occupancy=random.random(), bfactor=random.random()*10,
                        altloc=random.choice(uppercase), rmin=random.random()*2,
                        epsilon=random.random()/2, rmin14=random.random()*2,
                        epsilon14=random.random()/2)
        atom.xx, atom.xy, atom.xz = (random.random()*100-50 for i in range(3))
        atom.number = random.randint(1, 100)
        atom.vx, atom.vy, atom.vz = (random.random()*100-50 for i in range(3))
        atom.multipoles = np.random.rand(10) * 10

        fobj = BytesIO()
        pickle.dump(atom, fobj)
        fobj.seek(0)
        unpickled = pickle.load(fobj)

        self.assertIsInstance(unpickled, pmd.Atom)
        self._equal_atoms(unpickled, atom)

    def test_bond_serialization(self):
        """ Tests the serialization of Bond """
        struct = utils.create_random_structure(True)
        bond = struct.bonds[0]
        fobj = BytesIO()
        pickle.dump(bond, fobj)
        fobj.seek(0)
        unpickled = pickle.load(fobj)

        self.assertIsInstance(unpickled, pmd.Bond)
        self.assertEqual(unpickled.order, bond.order)
        self.assertEqual(unpickled.qualitative_type, bond.qualitative_type)

    def test_bondtype_serialization(self):
        """ Tests the serialization of BondType """
        struct = utils.create_random_structure(True)
        bt = struct.bond_types[0]

        fobj = BytesIO()
        pickle.dump(bt, fobj)
        fobj.seek(0)

        unpickled = pickle.load(fobj)

        self.assertEqual(unpickled, bt)
        self.assertIsNot(unpickled, bt)

    def test_residue_serialization(self):
        """ Tests the serialization of Residue """
        struct = utils.create_random_structure(parametrized=True)
        res = struct.residues[0]

        fobj = BytesIO()
        pickle.dump(res, fobj)
        fobj.seek(0)
        unpickled = pickle.load(fobj)

        self.assertEqual(len(res.atoms), len(unpickled.atoms))
        for a1, a2 in zip(res, unpickled):
            self._equal_atoms(a1, a2)
            self.assertIs(a1.residue, res)
            self.assertIs(a2.residue, unpickled)

    def test_structure_serialization(self):
        """ Tests the serialization of Structure """
        structure = utils.create_random_structure(parametrized=True)
        # Make sure we copy over exclusions
        structure.atoms[0].exclude(structure.atoms[10])
        fobj = BytesIO()
        pickle.dump(structure, fobj)
        fobj.seek(0)
        unpickled = pickle.load(fobj)

        self._compare_structures(unpickled, structure)

    def test_structure_serialization_with_bond_types(self):
        """Tests the serialization of Structure with extra bond attributes."""
        structure = utils.create_random_structure(parametrized=True)
        # Add some bond types
        structure.bonds[0].order = 2.0
        structure.bonds[0].qualitative_type = QualitativeBondType.AROMATIC
        fobj = BytesIO()
        pickle.dump(structure, fobj)
        fobj.seek(0)
        unpickled = pickle.load(fobj)
        self._compare_structures(unpickled, structure)
        self.assertEqual(unpickled.bonds[0].order, 2.0)
        self.assertEqual(unpickled.bonds[0].qualitative_type, QualitativeBondType.AROMATIC)

    def test_fortran_format_serialization(self):
        """ Tests the serialization of FortranFormat """
        fmt = pmd.amber.FortranFormat('8I10')
        unpickled = pickle.loads(pickle.dumps(fmt))

        self.assertEqual(fmt.format, unpickled.format)
        self.assertEqual(fmt.strip_strings, unpickled.strip_strings)
        self.assertIs(fmt.type, unpickled.type)
        self.assertEqual(fmt.nitems, unpickled.nitems)
        self.assertEqual(fmt.itemlen, unpickled.itemlen)
        self.assertEqual(fmt.fmt, unpickled.fmt)

    def test_amberformat_serialization(self):
        """ Tests the serialization of AmberFormat """
        amber = pmd.load_file(utils.get_fn('cSPCE.mdl'))
        unpickled = pickle.loads(pickle.dumps(amber))

        self.assertEqual(set(amber.parm_data.keys()), set(unpickled.parm_data.keys()))
        self.assertEqual(amber.flag_list, unpickled.flag_list)
        self.assertEqual(set(amber.formats.keys()), set(unpickled.formats.keys()))
        for k1 in amber.parm_data.keys():
            self.assertEqual(amber.parm_data[k1], unpickled.parm_data[k1])
            self.assertEqual(amber.formats[k1], unpickled.formats[k1])

        self.assertEqual(amber.charge_flag, unpickled.charge_flag)
        self.assertEqual(amber.version, unpickled.version)
        self.assertEqual(amber.name, unpickled.name)

    def test_amberparm_serialization(self):
        """ Tests the serialization of AmberParm """
        structure = pmd.load_file(utils.get_fn('ash.parm7'))
        unpickled = pickle.loads(pickle.dumps(structure))
        self.assertFalse(structure.unknown_functional)
        self.assertFalse(structure.unknown_functional)

        self._compare_structures(unpickled, structure)

        self.assertEqual(set(structure.parm_data.keys()), set(unpickled.parm_data.keys()))
        self.assertEqual(structure.flag_list, unpickled.flag_list)
        self.assertEqual(set(structure.formats.keys()), set(unpickled.formats.keys()))
        for k1 in structure.parm_data.keys():
            self.assertEqual(structure.parm_data[k1], unpickled.parm_data[k1])
            self.assertEqual(structure.formats[k1], unpickled.formats[k1])

        self.assertEqual(structure.charge_flag, unpickled.charge_flag)
        self.assertEqual(structure.version, unpickled.version)
        self.assertEqual(structure.name, unpickled.name)
        self.assertIs(pmd.amber.AmberParm, type(unpickled))

        for key in 'pointers LJ_types LJ_radius LJ_depth'.split():
            self.assertEqual(hasattr(structure, key),
                             hasattr(unpickled, key))
            if hasattr(structure, key):
                self.assertEqual(getattr(structure, key), getattr(unpickled, key))

        # Now check that unknown_functional gets properly deserialized
        structure.unknown_functional = True
        self.assertTrue(pickle.loads(pickle.dumps(structure)).unknown_functional)

    def test_chamberparm_serialization(self):
        """ Tests the serialization of ChamberParm """
        structure = pmd.load_file(utils.get_fn('ala_ala_ala.parm7'),
                                  utils.get_fn('ala_ala_ala.rst7'))
        unpickled = pickle.loads(pickle.dumps(structure))

        self._compare_structures(unpickled, structure)

        self.assertEqual(set(structure.parm_data.keys()), set(unpickled.parm_data.keys()))
        self.assertEqual(structure.flag_list, unpickled.flag_list)
        self.assertEqual(set(structure.formats.keys()), set(unpickled.formats.keys()))
        for k1 in structure.parm_data.keys():
            self.assertEqual(structure.parm_data[k1], unpickled.parm_data[k1])
            self.assertEqual(structure.formats[k1], unpickled.formats[k1])

        self.assertEqual(structure.charge_flag, unpickled.charge_flag)
        self.assertEqual(structure.version, unpickled.version)
        self.assertEqual(structure.name, unpickled.name)
        self.assertIs(pmd.amber.ChamberParm, type(unpickled))

    def test_amoebaparm_serialization(self):
        """ Tests the serialization of AmoebaParm """
        structure = pmd.load_file(utils.get_fn('nma.parm7'),
                                  utils.get_fn('nma.rst7'))
        unpickled = pickle.loads(pickle.dumps(structure))

        self._compare_structures(unpickled, structure)

        self.assertEqual(set(structure.parm_data.keys()), set(unpickled.parm_data.keys()))
        self.assertEqual(structure.flag_list, unpickled.flag_list)
        self.assertEqual(set(structure.formats.keys()), set(unpickled.formats.keys()))
        for k1 in structure.parm_data.keys():
            self.assertEqual(structure.parm_data[k1], unpickled.parm_data[k1])
            self.assertEqual(structure.formats[k1], unpickled.formats[k1])

        self.assertEqual(structure.charge_flag, unpickled.charge_flag)
        self.assertEqual(structure.version, unpickled.version)
        self.assertEqual(structure.name, unpickled.name)
        self.assertIs(pmd.amber.AmoebaParm, type(unpickled))

    def test_pdb_serialization(self):
        """ Tests the serialization of a parsed PDB file """
        structure = pmd.load_file(utils.get_fn('4lzt.pdb'))
        unpickled = pickle.loads(pickle.dumps(structure))

        self._compare_structures(unpickled, structure)

        # Check metadata
        for key in ('experimental', 'journal', 'authors', 'keywords', 'doi',
                    'pmid', 'journal_authors', 'volume', 'title', 'year',
                    'resolution', 'related_entries', 'space_group'):
            self.assertEqual(getattr(structure, key), getattr(unpickled, key))

    def test_pdbtraj_serialization(self):
        """ Tests the serialization of a Structure with trajectory """
        structure = pmd.load_file(utils.get_fn('2koc.pdb'))
        unpickled = pickle.loads(pickle.dumps(structure))

        self._compare_structures(unpickled, structure)

        # Check metadata
        for key in ('experimental', 'journal', 'authors', 'keywords', 'doi',
                    'pmid', 'journal_authors', 'volume', 'title', 'year',
                    'resolution', 'related_entries'):
            self.assertEqual(getattr(structure, key), getattr(unpickled, key))

        self.assertGreater(structure.get_coordinates().shape[0], 1)

    def test_parm_velocities_serialization(self):
        """ Tests the serialization of a Structure with velocities """
        structure = pmd.load_file(utils.get_fn('tip4p.parm7'), utils.get_fn('tip4p.rst7'))
        unpickled = pickle.loads(pickle.dumps(structure))
        self._compare_structures(unpickled, structure)

    @unittest.skipUnless(HAS_GROMACS, "Cannot run GROMACS tests without GROMACS")
    def test_gromacstop_serialization(self):
        """ Tests the serialization of a GromacsTopologyFile """
        structure = pmd.load_file(os.path.join(utils.get_fn('03.AlaGlu'), 'topol.top'),
                                  xyz=os.path.join(utils.get_fn('03.AlaGlu'), 'conf.gro'))
        unpickled = pickle.loads(pickle.dumps(structure))
        self._compare_structures(unpickled, structure)
        self.assertEqual(structure.defaults, unpickled.defaults)
        self._compare_parametersets(structure.parameterset, unpickled.parameterset)

    @unittest.skipUnless(HAS_GROMACS, "Cannot run GROMACS tests without GROMACS")
    def test_gromacscharmm_serialization(self):
        """ Tests the serialization of a CHARMM FF Gromacs topology """
        structure = pmd.load_file(utils.get_fn('1aki.charmm27.solv.top'))
        unpickled = pickle.loads(pickle.dumps(structure))
        self._compare_structures(unpickled, structure)
        self.assertEqual(structure.defaults, unpickled.defaults)
        self._compare_parametersets(structure.parameterset, unpickled.parameterset)

    def test_charmmpsf_serialization(self):
        """ Tests the serialization of a CHARMM PSF file """
        structure = pmd.load_file(utils.get_fn('ala_ala_ala.psf'))
        unpickled = pickle.loads(pickle.dumps(structure))
        self._compare_structures(unpickled, structure)

    def _compare_structures(self, unpickled, structure):

        self.assertEqual(len(unpickled.residues), len(structure.residues))
        for r1, r2 in zip(unpickled.residues, structure.residues):
            self.assertEqual(len(r1), len(r2))
            self.assertEqual(r1.idx, r2.idx)
            for a1, a2 in zip(r1, r2):
                self._equal_atoms(a1, a2)

        def cmp_alists(alist1, alist2):
            self.assertEqual(len(alist1), len(alist2))
            for a1, a2 in zip(alist1, alist2):
                self._equal_atoms(a1, a2)

        for a1, a2 in zip(unpickled, structure):
            self._equal_atoms(a1, a2)
            self.assertEqual(a1.idx, a2.idx)
            self.assertEqual(len(a1.bonds), len(a2.bonds))
            self.assertEqual(len(a1.angles), len(a2.angles))
            self.assertEqual(len(a1.dihedrals), len(a2.dihedrals))
            self.assertEqual(len(a1.impropers), len(a2.impropers))
            cmp_alists(a1.bond_partners, a2.bond_partners)
            cmp_alists(a1.angle_partners, a2.angle_partners)
            cmp_alists(a1.dihedral_partners, a2.dihedral_partners)
            cmp_alists(a1.tortor_partners, a2.tortor_partners)
            cmp_alists(a1.exclusion_partners, a2.exclusion_partners)

        # Check coordinates
        if structure.get_coordinates() is None:
            self.assertIs(unpickled.get_coordinates(), None)
        else:
            np.testing.assert_equal(structure.get_coordinates(), unpickled.get_coordinates())
        # Check unit cell
        if structure.box is None:
            self.assertIs(unpickled.box, None)
            self.assertIs(unpickled.box_vectors, None)
        else:
            np.testing.assert_equal(structure.box, unpickled.box)
            self.assertEqual(structure.box_vectors, unpickled.box_vectors)

        # Check velocities
        if structure.velocities is None:
            self.assertIs(unpickled.velocities, None)
        else:
            np.testing.assert_equal(structure.velocities, unpickled.velocities)

        # Check nrexcl and combining_rule
        self.assertEqual(structure.nrexcl, unpickled.nrexcl)
        self.assertEqual(structure.combining_rule, unpickled.combining_rule)

        # Make sure all of the type arrays are equivalent
        def cmp_type_arrays(arr1, arr2):
            self.assertEqual(len(arr1), len(arr2))
            for x1, x2 in zip(arr1, arr2):
                self.assertEqual(x1, x2)

        cmp_type_arrays(structure.bond_types, unpickled.bond_types)
        cmp_type_arrays(structure.angle_types, unpickled.angle_types)
        cmp_type_arrays(structure.dihedral_types, unpickled.dihedral_types)
        cmp_type_arrays(structure.improper_types, unpickled.improper_types)
        cmp_type_arrays(structure.urey_bradley_types, unpickled.urey_bradley_types)
        cmp_type_arrays(structure.rb_torsion_types, unpickled.rb_torsion_types)
        cmp_type_arrays(structure.cmap_types, unpickled.cmap_types)
        cmp_type_arrays(structure.trigonal_angle_types, unpickled.trigonal_angle_types)
        cmp_type_arrays(structure.out_of_plane_bend_types, unpickled.out_of_plane_bend_types)
        cmp_type_arrays(structure.stretch_bend_types, unpickled.stretch_bend_types)
        cmp_type_arrays(structure.torsion_torsion_types, unpickled.torsion_torsion_types)
        cmp_type_arrays(structure.pi_torsion_types, unpickled.pi_torsion_types)
        cmp_type_arrays(structure.adjust_types, unpickled.adjust_types)

        # Make sure all of the connectivity arrays are equivalent
        def cmp_top_arrays(arr1, arr2):
            self.assertEqual(len(arr1), len(arr2))
            for t1, t2 in zip(arr1, arr2):
                self.assertIs(type(t1), type(t2))
                atoms = [attr for attr in dir(t1) if attr.startswith('atom')]
                for a in atoms:
                    self._equal_atoms(getattr(t1, a), getattr(t2, a))
                if hasattr(t1, 'type'):
                    self.assertEqual(t1.type, t2.type)

        cmp_top_arrays(structure.bonds, unpickled.bonds)
        cmp_top_arrays(structure.angles, unpickled.angles)
        cmp_top_arrays(structure.dihedrals, unpickled.dihedrals)
        cmp_top_arrays(structure.impropers, unpickled.impropers)
        cmp_top_arrays(structure.urey_bradleys, unpickled.urey_bradleys)
        cmp_top_arrays(structure.rb_torsions, unpickled.rb_torsions)
        cmp_top_arrays(structure.cmaps, unpickled.cmaps)
        cmp_top_arrays(structure.trigonal_angles, unpickled.trigonal_angles)
        cmp_top_arrays(structure.out_of_plane_bends, unpickled.out_of_plane_bends)
        cmp_top_arrays(structure.pi_torsions, unpickled.pi_torsions)
        cmp_top_arrays(structure.stretch_bends, unpickled.stretch_bends)
        cmp_top_arrays(structure.torsion_torsions, unpickled.torsion_torsions)
        cmp_top_arrays(structure.chiral_frames, unpickled.chiral_frames)
        cmp_top_arrays(structure.multipole_frames, unpickled.multipole_frames)
        cmp_top_arrays(structure.adjusts, unpickled.adjusts)
        cmp_top_arrays(structure.groups, unpickled.groups)

    def _compare_parametersets(self, set1, set2):

        def cmp_lists(l1, l2, reversible=False):
            self.assertEqual(len(l1), len(l2))
            self.assertEqual(set(l1.keys()), set(l2.keys()))
            for k in l1:
                self.assertEqual(l1[k], l2[k])
                if reversible:
                    if pmd.NoUreyBradley is l1[k] or pmd.NoUreyBradley is l2[k]:
                        self.assertIs(l1[k], pmd.NoUreyBradley)
                        self.assertIs(l2[k], pmd.NoUreyBradley)
                    self.assertIs(l1[k], l1[tuple(reversed(k))])
                    self.assertIs(l2[k], l2[tuple(reversed(k))])

        # Now compare all of the parameter lists
        cmp_lists(set1.bond_types, set2.bond_types, True)
        cmp_lists(set1.angle_types, set2.angle_types, True)
        cmp_lists(set1.dihedral_types, set2.dihedral_types, True)
        cmp_lists(set1.improper_types, set2.improper_types)
        cmp_lists(set1.improper_periodic_types, set2.improper_periodic_types)
        cmp_lists(set1.cmap_types, set2.cmap_types, True)
        cmp_lists(set1.atom_types, set2.atom_types)
        cmp_lists(set1.urey_bradley_types, set2.urey_bradley_types, True)
        cmp_lists(set1.nbfix_types, set2.nbfix_types)
        self.assertEqual(set1.combining_rule, set2.combining_rule)



""" Standard residue templates for biomolecular residues """

import json
from collections import Counter
from io import TextIOWrapper
try:
    from functools import cache
except ImportError:
    # Patch to support a `cache` decorator in Python 3.8 equivalent to functools.cache in Python 3.9+
    from functools import lru_cache
    cache = lru_cache(maxsize=None)
from typing import Mapping, Dict, Any, List, Set, TextIO
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pathlib import Path
import os
from ..amber.offlib import AmberOFFLibrary
from .residue import ResidueTemplate
from ..topologyobjects import Atom, QualitativeBondType, Hybridization
from ..periodic_table import AtomicNum

__all__ = ['StandardBiomolecularResidues']

StandardBiomolecularResidues = AmberOFFLibrary.parse(
    os.path.join(os.path.split(__file__)[0], 'data', 'standard_residues.lib')
)

def load_ccd_residue_templates(fileobj: TextIO) -> Mapping[str, ResidueTemplate]:
    """Loads a biopolymer residue database from a dictionary prepared from the chemical component dictionary

    Parameters
    ----------
    fileobj: file-like
        The file object to read the CCD residue templates from

    Returns
    -------
    The dictionary mapping residue template name with the ResidueTemplate object
    """
    residues = dict()
    all_residue_data = json.load(fileobj)

    for single_residue in all_residue_data:
        residues[single_residue["name"]] = _process_single_residue(single_residue)

    # Augment common residues from the Amber-standard residues
    extra_residues = set(StandardBiomolecularResidues.keys()) - set(residues.keys())
    residues.update({name: StandardBiomolecularResidues[name] for name in extra_residues})
    return residues


def _process_single_residue(data: Dict[str, Any]) -> ResidueTemplate:
    """Processes data from a single residue and builds a template from it."""
    # In the CCD, a "leaving" atom is one that is "removed" when a residue forms a polymeric chain
    # with an adjacent residue. We adopt the convention here that the 'head' and 'tail' of the polymeric
    # residues are the first and last atoms attached to those that are "leaving"
    #
    # It is harder to determine whether a single linkage point is a head or tail. That is done by
    # <insert heuristic algorithm here>
    template = ResidueTemplate(data["name"])
    _add_atoms(data["atoms"], template)
    _add_bonds(data["bonds"], template)
    _assign_head_tail(template, {atom["name"] for atom in data["atoms"] if atom["leaving"]}, data["type"])
    return template

def _add_atoms(atoms: List[Dict[str, Any]], template: ResidueTemplate) -> None:
    """Adds atoms to a residue template"""
    for atom in atoms:
        name = atom["name"]
        symbol = atom["symbol"]
        formal_charge = atom["formal_charge"]
        aromatic = atom["aromatic"]

        hybridization_value = atom.get("hybridization", None)
        hybridization = Hybridization(hybridization_value) if hybridization_value is not None else None

        atom = Atom(
            atomic_number=AtomicNum[symbol.title()], type=symbol, name=name, formal_charge=formal_charge,
            hybridization=hybridization, aromatic=aromatic
        )
        template.add_atom(atom)

def _add_bonds(bonds: List[Dict[str, Any]], template: ResidueTemplate) -> None:
    """Adds bonds to a template"""
    atom_map = {atom.name: atom for atom in template.atoms}
    _order_map = {
        ("SING", False): 1.0, ("SING", True): 1.5, ("DOUB", False): 2.0, ("DOUB", True): 1.5, ("TRIP", False): 3.0,
        ("TRIP", True): 2.5
    }
    _qualitative_map = {
        ("SING", False): QualitativeBondType.SINGLE, ("DOUB", False): QualitativeBondType.DOUBLE,
        ("SING", True): QualitativeBondType.AROMATIC, ("DOUB", True): QualitativeBondType.AROMATIC,
        ("TRIP", False): QualitativeBondType.TRIPLE, ("TRIP", True): QualitativeBondType.AROMATIC,
    }
    for bond in bonds:
        key = (bond["order"], bond["aromatic"])
        template.add_bond(
            atom_map[bond["name_a"]], atom_map[bond["name_b"]],
            order=_order_map[key], qualitative_type=_qualitative_map[key]
        )


def _head_or_tail(atom: Atom, residue_type: str) -> Literal["head", "tail", "neither", "maybe head", "maybe tail"]:
    if 'peptide' in residue_type.lower():
        if atom.atomic_number == AtomicNum["N"]:
            return "head" if atom.name == "N" else "maybe head"
        if atom.atomic_number == AtomicNum["C"]:
            return "tail" if atom.name == "C" else "maybe tail"
        return "neither"

    if 'rna' in residue_type.lower() or 'dna' in residue_type.lower():
        if atom.atomic_number == AtomicNum["P"]:
            return "head" if atom.name == "P" else "maybe head"
        elif atom.atomic_number == AtomicNum["O"]:
            return "tail" if atom.name == "O3'" else "maybe tail"
        return "neither"

    return "neither"

def _assign_head_tail(template: ResidueTemplate, removable_atoms: Set[str], residue_type: str) -> None:
    """Assigns the head and tail of a residue based on the atoms that can be removed"""
    atoms_joined_to_removable = set()
    for atom in template.atoms:
        if atom.name not in removable_atoms:
            continue
        for partner in atom.bond_partners:
            if partner.name in removable_atoms:
                continue
            atoms_joined_to_removable.add(partner)

    sorted_connectivity_atoms = sorted(atoms_joined_to_removable)
    head_tail_assignments = [_head_or_tail(atom, residue_type) for atom in sorted_connectivity_atoms]
    head_tail_counts = Counter(head_tail_assignments)
    assert head_tail_counts.get("head", 0) <= 1, "Too many heads. Check residue template file"
    assert head_tail_counts.get("tail", 0) <= 1, "Too many tails. Check residue template file"

    # Assign the head. In a spot-check of the residues where there was no atom that should clearly be assigned as
    # the head, the "first" atom in the sequence that was tagged as "maybe head" was always the correct choice.
    # Use that strategy
    if head_tail_counts.get("head", 0) == 1:
        for atom, label in zip(sorted_connectivity_atoms, head_tail_assignments):
            if label == "head":
                template.head = atom
                break
    else:
        for atom, label in zip(sorted_connectivity_atoms, head_tail_assignments):
            if label == "maybe head":
                # Stop at the first
                template.head = atom
                break

    # Assign the tail. In a spot-check of the residues where there was no atom that should clearly be assigned as
    # the head, the "last" atom in the sequence that was tagged as "maybe tail" was always the correct choice.
    # Use that strategy
    if head_tail_counts.get("tail", 0) == 1:
        for atom, label in zip(sorted_connectivity_atoms, head_tail_assignments):
            if label == "tail":
                template.tail = atom
                break
    else:
        for atom, label in zip(sorted_connectivity_atoms, head_tail_assignments):
            if label == "maybe tail":
                template.tail = atom
                # Do not stop at the first in case there's another one later.
                # Reversing would be faster, but optimization is not needed

    # Assign the remaining connectivity atoms that are neither head nor tail
    for atom in sorted_connectivity_atoms:
        if atom is not template.head and atom is not template.tail:
            template.connections.append(atom)


_CACHED_RESIDUE_TEMPLATE_LIBRARY = None

@cache
def get_standard_residue_template_library() -> Dict[str, ResidueTemplate]:
    path = Path(__file__).parent / "data" / "ccd_residue_templates.json"
    with path.open("r") as ifh:
        return load_ccd_residue_templates(ifh)


@cache
def get_nonstandard_ccd_residues() -> Dict[str, ResidueTemplate]:
    import gzip
    path = Path(__file__).parent / "data" / "nonstandard_ccd_residue_templates.json.gz"
    with TextIOWrapper(gzip.open(path, "rb")) as ifh:
        return load_ccd_residue_templates(ifh)



"""
Tools to aid in input/output within the parmed package
"""

__all__ = ['genopen']

import os
from io import TextIOWrapper, BytesIO
from pathlib import Path
from typing import Union
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

from ..constants import DEFAULT_ENCODING

def genopen(name: Union[Path, str], mode: str = 'r'):
    """
    Opens a file, automatically detecting compression schemes by filename
    extension. Note, these files are opened in a way that *always* returns a
    string. This is an important distinction in Python 3 where many file-like
    objects return bytes instead of strings. This is detected and handled
    properly so that the object returned from this function is always
    string-based (so you cannot write or read bytes directly from a file opened
    via ``genopen``).

    This routine also recognizes URLs and will read remote files when given a
    URL starting with either http:// or https://. Like with standard local file
    names, compression is automatically detected by filename extension, and both
    gzip and bzip2 files are supported.

    Parameters
    ----------
    name : str or Path
        Name of the file to open or URL to a remote file to access
    mode : str, optional
        Whether to open the file to 'r'ead, 'w'rite, or 'a'ppend. Default is 'r'

    Returns
    -------
    file : file-like
        A file-like object in the requested mode

    Notes
    -----
    Python's BZ2File does not support writing to ``append`` mode (mode='a'), so
    it is faked here. The entire file contents are read into memory and then
    written into a 'new' file with the same name as the original. As such, it is
    noticeably slower and more resource-intensive (particularly for large files)
    than using gzipped files.
    """
    if mode not in ['w', 'r', 'a']:
        raise ValueError('open mode must be "w", "r", or "a"')

    # Handle arbitrary online files. file:// is just an alias for a local file
    is_url = False
    if name.startswith('file:///'):
        name = name[7:]
    elif name.startswith('http://') or name.startswith('https://') or name.startswith('ftp://'):
        is_url = True
        if mode in ['w', 'a']:
            raise ValueError('Cannot write or append a webpage')
        try:
            open_url = urlopen(name)
        except (HTTPError, URLError) as err:
            raise IOError(f"Could not open {name}") from err

    if name.endswith('.bz2'):
        import bz2
        # BZ2File cannot open in append mode, so we have to fake it. Read the
        # entire existing contents into memory, open a new file, write the
        # contents back, and return the file that is now open for writing
        if mode == 'a':
            tmp = BytesIO()
            if os.path.exists(name):
                with bz2.BZ2File(name, 'rb') as f:
                    tmp.write(f.read())
                tmp.seek(0)
            f = bz2.BZ2File(name, 'wb')
            f.write(tmp.read())
            return TextIOWrapper(f)
        # If it is a URL, just pass in the urlopen object as a filename
        if is_url:
            name = open_url
        return TextIOWrapper(bz2.BZ2File(name, mode+'b'))
    elif name.endswith('.gz'):
        import gzip
        if is_url:
            return TextIOWrapper(gzip.GzipFile(fileobj=open_url, mode='r'))
        else:
            return TextIOWrapper(gzip.open(name, mode+'b'))

    if is_url:
        return TextIOWrapper(open_url)
    else:
        return open(name, mode, encoding=DEFAULT_ENCODING)



"""
Tests the functionality in the tinker subpackage
"""
import numpy as np
from utils import get_fn
import unittest
import parmed as pmd
from parmed.tinker import parameterfile, system, tinkerfiles

class TestTinkerFiles(unittest.TestCase):

    def test_parameter_file(self):
        """ Tests parsing TINKER parameter files """
        param = parameterfile.AmoebaParameterSet(get_fn('amoeba09.prm'))
        attributes = {'angle-sextic': 2.2e-08, 'direct-13-scale': 1.0,
                      'radiustype': 'R-MIN', 'mpole-14-scale': 0.4,
                      'direct-12-scale': 1.0, 'mpole-12-scale': 0.0,
                      'mutual-13-scale': 1.0, 'mpole-13-scale': 0.0,
                      'opbend-pentic': -7e-07, 'vdw-13-scale': 0.0,
                      'angle-cubic': -0.014, 'mutual-12-scale': 1.0,
                      'angle-pentic': -7e-07, 'vdw-14-scale': 1.0,
                      'polarization': 'MUTUAL', 'polar-13-scale': 0.0,
                      'forcefield': 'AMOEBA-2009', 'radiussize': 'DIAMETER',
                      'bond-quartic': 3.793125, 'vdwtype': 'BUFFERED-14-7',
                      'polar-12-scale': 0.0, 'mutual-14-scale': 1.0,
                      'mutual-11-scale': 1.0, 'direct-14-scale': 1.0,
                      'torsionunit': 0.5, 'polar-15-scale': 1.0, 'radiusrule':
                      'CUBIC-MEAN', 'dielectric': 1.0, 'polar-14-intra': 0.5,
                      'vdw-12-scale': 0.0, 'direct-11-scale': 0.0, 'opbendtype':
                      'ALLINGER', 'opbend-quartic': 5.6e-05, 'polar-14-scale':
                      1.0, 'opbend-sextic': 2.2e-08, 'vdw-15-scale': 1.0,
                      'bond-cubic': -2.55, 'opbend-cubic': -0.014,
                      'angle-quartic': 5.6e-05, 'mpole-15-scale': 0.8,
                      'epsilonrule': 'HHG'}
        for key in attributes:
            self.assertEqual(param.attributes[key], attributes[key])

        self.assertEqual(len(param.angles), 18)
        self.assertAlmostEqual(param.angles['27-31-29'].k, 48.2)
        self.assertAlmostEqual(param.angles['27-31-29'].theteq, 109.5)
        self.assertAlmostEqual(param.angles['27-31-29'].theteq2, 110.2)
        self.assertAlmostEqual(param.angles['27-31-29'].theteq3, 111.0)

        self.assertEqual(len(param.bonds), 13)
        self.assertAlmostEqual(param.bonds['18-19'].k, 1072.0)
        self.assertAlmostEqual(param.bonds['18-19'].req, 1.1874)

        self.assertEqual(len(param.dihedrals), 150)
        self.assertEqual(param.dihedrals['41-40-40-51'].k, [0, 0, 0.299])
        self.assertEqual(param.dihedrals['41-40-40-51'].phase, [0, 180.0, 0])
        self.assertEqual(param.dihedrals['41-40-40-51'].periodicity, [1, 2, 3])

        self.assertEqual(len(param.opbends), 35)
        self.assertAlmostEqual(param.opbends['60-62-0-0'].k, 107.9)

        self.assertEqual(len(param.torsion_torsions), 0)

        self.assertEqual(len(param.urey_bradleys), 1)
        self.assertEqual(param.urey_bradleys['35-34-35'].k, -7.6)
        self.assertEqual(param.urey_bradleys['35-34-35'].req, 1.5326)

        self.assertEqual(len(param.multipoles), 17)
        self.assertTrue(not any(param.multipoles['2-0-0'].potential_terms))

    def test_analout(self):
        """ Tests parsing Amber-style Tinker analout files """
        analout = system.TinkerAnalout(get_fn('jac.analout'))
        self.assertEqual(len(analout.atom_list), 23558)
        self.assertEqual(len(analout.angle_list), 11584)
        self.assertEqual(len(analout.atom_inter_flags), 23)
        self.assertEqual(len(analout.atom_list), 23558)
        self.assertEqual(len(analout.bond_list), 16569)
        self.assertEqual(len(analout.dipole_list), 23558)
        self.assertEqual(len(analout.multipole_list), 23558)
        self.assertEqual(len(analout.oopbend_list), 1566)
        self.assertEqual(len(analout.pair12_list), 16569)
        self.assertEqual(len(analout.pair13_list), 11584)
        self.assertEqual(len(analout.pair14_list), 6556)
        self.assertEqual(len(analout.pair15_list), 7476)
        self.assertEqual(len(analout.pitors_list), 292)
        self.assertEqual(len(analout.pointers), 23)
        self.assertEqual(len(analout.stretchbend_list), 4031)
        self.assertEqual(len(analout.torangle_list), 6701)
        self.assertEqual(len(analout.tortor_list), 147)
        self.assertEqual(len(analout.ureybrad_list), 7023)

        # Sample some random attributes and assume this is comprehensive
        self.assertEqual(analout.multipole_list[5].definition, 'Z-then-X')
        self.assertEqual(analout.multipole_list[5].frame, [1, 2])
        self.assertEqual(analout.multipole_list[5].moment,
                         [0.2124, 0.0, 0.0, -0.1249, 0.03622, 0.0, -0.01437,
                          0.0, 0.0, -0.02185])

        self.assertEqual(analout.pitors_list[5].atom1.type, 9)
        self.assertEqual(len(analout.pair12_list), 16569)

    def test_xyz(self):
        """ Tests parsing Tinker XYZ files """
        self.assertTrue(tinkerfiles.XyzFile.id_format(get_fn('nma.xyz')))
        self.assertTrue(tinkerfiles.XyzFile.id_format(get_fn('2igd_924wat.xyz')))
        xyz = tinkerfiles.XyzFile(get_fn('nma.xyz'))
        np.testing.assert_allclose(xyz.box, [30.735, 30.876, 28.485, 90.0, 90.0, 90.0])
        self.assertEqual(len(xyz.atoms), 2466)
        self.assertEqual(xyz.atoms[0].name, 'C')
        self.assertEqual(xyz.atoms[0].type, '221')
        self.assertEqual(len(xyz.atoms[0].bond_partners), 4)
        self.assertEqual(xyz.atoms[-1].name, 'H')
        self.assertEqual(xyz.atoms[-1].atomic_number, 1)
        self.assertEqual(xyz.atoms[-1].type, '248')
        xyz = pmd.load_file(get_fn('2igd_924wat.xyz'), get_fn('2igd_924wat.pdb'))
        pdb = pmd.load_file(get_fn('2igd_924wat.pdb'))
        xyz2 = pmd.load_file(get_fn('2igd_924wat.xyz'))
        self.assertEqual(len(pdb.atoms), len(xyz.atoms))
        self.assertEqual(len(pdb.residues), len(xyz.residues))
        for r1, r2 in zip(pdb.residues, xyz.residues):
            self.assertEqual(len(r1), len(r2))
            self.assertEqual(r1.chain, r2.chain)
            self.assertEqual(r1.name, r2.name)
            self.assertEqual(r1.insertion_code, r2.insertion_code)
        # Make sure NA ions are atomic number of sodium
        for atom in xyz.view['NA',:].atoms:
            self.assertEqual(atom.atomic_number, pmd.periodic_table.AtomicNum['Na'])
        # Now make sure that NA ions are atomic number of sodium even if we
        # don't load a PDB file
        for atom in xyz2.view[:,'Na+']:
            self.assertEqual(atom.atomic_number, pmd.periodic_table.AtomicNum['Na'])

    def test_dyn(self):
        """ Tests parsing Tinker DYN files """
        dyn = tinkerfiles.DynFile(get_fn('nma.dyn'))
        self.assertEqual(dyn.natom, 2466)
        for attr in ('accelerations', 'old_accelerations', 'box', 'positions',
                     'velocities'):
            self.assertTrue(hasattr(dyn, attr))

        for x, y in zip(dyn.positions[10],
                [-0.1099425448789507, -1.83499212341286, 6.089155631551154]):
            self.assertAlmostEqual(x, y)



""" XML file parsing utilities for OpenMM serialized objects """
import re
from contextlib import closing
from io import StringIO
import numpy as np
from ..vec3 import Vec3
from ..formats.registry import FileFormatType
from ..geometry import box_vectors_to_lengths_and_angles
from .. import unit as u
from ..utils.decorators import needs_openmm
from ..utils.io import genopen

_xmlre = re.compile('<(.*?)/?>')

class XmlFile(metaclass=FileFormatType):
    """
    Wrapper for parsing OpenMM-serialized objects. Supports serialized State,
    System, Integrator, and ForceField objects.
    """

    @staticmethod
    def id_format(filename):
        """ Identifies the file type as an XML file

        Parameters
        ----------
        filename : str
            Name of the file to check format for

        Returns
        -------
        is_fmt : bool
            True if it is an XML format, False otherwise
        """
        with closing(genopen(filename, 'r')) as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                rematch = _xmlre.match(line)
                if not rematch:
                    return False
                stuff = rematch.groups()[0]
                if stuff[0] in '?!':
                    continue
                kind = stuff.split()[0]
                if kind in ('System', 'State', 'ForceField', 'Integrator'):
                    return True
                return False

    @staticmethod
    @needs_openmm
    def parse(filename):
        """
        Parses XML file and returns deserialized object. The return value
        depends on the serialized object, summarized below

            - System : returns openmm.System
            - State : returns openmm.State
            - Integrator : returns openmm.Integrator subclass
            - ForceField : returns openmm.app.ForceField

        Parameters
        ----------
        filename : str or file-like
            The file name or file object containing the XML-serialized object

        Returns
        -------
        obj : System, State, Integrator, or ForceField
            The deserialized object

        Notes
        -----
        OpenMM requires the entire contents of this file read into memory. As a
        result, this function may require a significant amount of memory.
        """
        import openmm as mm
        from openmm import app
        if isinstance(filename, str):
            with closing(genopen(filename, 'r')) as f:
                contents = f.read()
        else:
            contents = filename.read()
        # ForceField is not handled by XmlSerializer
        if '<ForceField' in contents:
            obj = StringIO()
            obj.write(contents)
            obj.seek(0)
            return app.ForceField(obj)

        obj = mm.XmlSerializer.deserialize(contents)
        if isinstance(obj, mm.State):
            return _OpenMMStateContents(obj)
        return obj

class _OpenMMStateContents:
    """
    A container that holds all of the information present in the OpenMM State
    object that gets deserialized. This is an internal class that is only
    intended to be *instantiated* by XmlFile.parse.

    Parameters
    ----------
    state : :class:`openmm.State`
        OpenMM State containing relevant state information

    Attributes
    ----------
    coordinates : np.ndarray with shape(1, natom, 3)
        Atomic coordinates. If not present in the State, just set to None. Units
        are Angstrom
    positions : list of Vec3
        Atomic coordinates of first frame with applied units
    velocities : np.ndarray with shape(1, natom, 3)
        Atomic velocities. If not present in the State, just set to None. Units
        are Angstrom/picosecond
    forces : np.ndarray with shape(1, natom, 3)
        Atomic forces. If not present in the State, just set to None. Units are
        kcal/mol/Angstrom
    energy : float
        Energy of this configuration in kcal/mol. If not specified, it is set to
        None
    time : float
        The time associated with this State in ps. If not specified, set to None
    """

    @staticmethod
    def _get_data(state, getter, unit, shape=None, **kwargs):
        try:
            stuff = getattr(state, getter)(**kwargs)
        except Exception:
            return None
        else:
            stuff = stuff.value_in_unit(unit)
            if shape is not None:
                stuff = stuff.reshape(shape)
            return stuff

    def __init__(self, state):
        self.coordinates = self._get_data(
            state, 'getPositions', u.angstrom, (1, -1, 3), asNumpy=True
        )
        self.velocities = self._get_data(
            state, 'getVelocities', u.angstrom/u.picosecond, (1, -1, 3), asNumpy=True
        )
        self.forces = self._get_data(
            state, 'getForces', u.kilocalorie_per_mole/u.angstrom, (1, -1, 3), asNumpy=True
        )
        self.energy = self._get_data(state, 'getPotentialEnergy', u.kilocalorie_per_mole)
        self.time = self._get_data(state, 'getTime', u.picosecond)
        box = self._get_data(state, 'getPeriodicBoxVectors', u.angstroms, (3, 3), asNumpy=True)
        if box is not None:
            leng, ang = box_vectors_to_lengths_and_angles(*box)
            leng = leng.value_in_unit(u.angstrom)
            ang = ang.value_in_unit(u.degree)
            self.box = np.array(list(leng) + list(ang))
        else:
            self.box = None

    @property
    def positions(self):
        return [Vec3(*xyz) for xyz in self.coordinates[0]]
